import click

from .host import (
    host_create,
    host_delete,
    host_list,
    host_rename,
    host_run,
    host_set,
    host_show,
)


# ------------------------------------------------------------------------------
# HOST Commands
# ------------------------------------------------------------------------------
@click.group(name="host", help="Command group for managing hosts")
def ssh_host():
    pass


# // Linking other sub-commands

ssh_host.add_command(host_create.cmd)
ssh_host.add_command(host_delete.cmd)
ssh_host.add_command(host_list.cmd)
ssh_host.add_command(host_set.cmd)
ssh_host.add_command(host_show.cmd)
ssh_host.add_command(host_rename.cmd)
ssh_host.add_command(host_run.cmd)



from typing import List

import click
from rich import box
from rich.console import Console
from rich.panel import Panel
from rich.table import Table

from sshtmux.sshm import SSH_Config, complete_ssh_group_names

# ------------------------------------------------------------------------------
# COMMAND: group show
# ------------------------------------------------------------------------------
SHORT_HELP = "Shows group details"
LONG_HELP = """
Display/Shows details from a group

Currently WIP - command will allow some styles in outputs...
"""

# Parameters help:
# ------------------------------------------------------------------------------


@click.command(name="show", short_help=SHORT_HELP, help=LONG_HELP)
@click.argument("name", shell_complete=complete_ssh_group_names)
@click.pass_context
def cmd(ctx, name):
    config: SSH_Config = ctx.obj

    if not config.check_group_by_name(name):
        print(f"Cannot show group '{name}', it is not defined in configuration!")
        ctx.exit(1)

    found_group = config.get_group_by_name(name)
    host_list: List[str] = []

    for host in found_group.hosts:
        if "hostname" in host.params:
            h_name = host.params["hostname"]
            h_port = (":" + host.params["port"]) if "port" in host.params else ""
            host_list.append(f"{host.name} ({h_name}{h_port})")
        else:
            host_list.append(host.name)

    pattern_list: List[str] = []
    for host in found_group.patterns:
        # hack to search via case insensitive info
        if "hostname" in host.params:
            h_name = host.params["hostname"]
            pattern_list.append(f"{host.name} ({h_name})")
        else:
            pattern_list.append(host.name)

    # New rich formatting
    table = Table(box=box.SQUARE, style="grey35")
    table.add_column("Group Parameter", no_wrap=True)
    table.add_column("Value")

    table.add_row("name", found_group.name, style="white")
    table.add_row("description", found_group.desc, style="grey50")
    table.add_row("info", Panel("\n".join(found_group.info)), style="grey50")
    table.add_row("hosts", Panel("\n".join(host_list)), style="white")
    table.add_row("patterns", Panel("\n".join(pattern_list)), style="cyan")

    console = Console()
    console.print(table)



import click

from .cmds import cmd_group, cmd_host, cmd_identity, cmd_snippets
from .cmds.cmd_group import group_list
from .cmds.cmd_host import host_list
from .main_tui import SSHTui
from .sshm import SSH_Config
from .version import VERSION

CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])

# ------------------------------------------------------------------------------
# COMMAND: sshm
# ------------------------------------------------------------------------------
MAIN_HELP = f"""
SSHTMux - Powerful SSH terminal manager. version {VERSION}

NOTE: As this will change your SSH config files, make backups before
using this software, as you might accidentally lose some configuration.
"""

# Parameters help:
STDOUT_HELP = "Send changed SSH config to STDOUT instead to original file, can be enabled with setting ENV variable (export SSHM_STDOUT=1)"
# ------------------------------------------------------------------------------


# In cases we want to have some execution without any sub-commands, instead of displaying help
# we can add "invoke_without_command=True" in a group decorator, to make function runnable directly
@click.group(context_settings=CONTEXT_SETTINGS, help=MAIN_HELP)
@click.option("--stdout", is_flag=True, envvar="SSHM_STDOUT", help=STDOUT_HELP)
@click.version_option(VERSION, message="SSHTMUX (sshm) - Version: %(version)s")
@click.pass_context
def cli(ctx: click.core.Context, stdout: bool):
    ctx.obj = SSH_Config(stdout=stdout).read().parse()


TUI_SHORT_HELP = "TUI Interface"
TUI_HELP = "Open TUI interface for interacting with SSH Configuration"


@click.command(name="tui", short_help=TUI_SHORT_HELP, help=TUI_HELP)
@click.pass_context
def tui_cmd(ctx: click.core.Context):
    SSHTui(ctx.obj).run()


# Link all commands to root command
# ------------------------------------------------------------------------------
# Top commands
cli.add_command(cmd_host.ssh_host)
cli.add_command(cmd_group.ssh_group)
cli.add_command(cmd_identity.generate)
cli.add_command(cmd_snippets.generate)
cli.add_command(tui_cmd)

# Top level aliases (groups --> group list, hosts --> host list, etc..)
cli.add_command(group_list.cmd, "groups")
cli.add_command(host_list.cmd, "hosts")



import importlib
import socket
from dataclasses import dataclass, field
from typing import Dict, Tuple

from rich.console import Console

from ..core.config import settings

console = Console()

DEBUG_STYLES = False


@dataclass
class SSH_Host:
    """Class for SSH host config structure"""

    name: str
    group: str
    type: str = "normal"
    info: list = field(default_factory=list)
    params: dict = field(default_factory=dict)

    inherited_params: list = field(default_factory=list)
    print_style: str = settings.sshtmux.SSHTMUX_HOST_STYLE

    def get_all_params(self) -> Dict[str, str]:
        """
        Method combines configured host parameters with all
        host inherited parameters, and returns them as dictionary
        """
        return {
            **self.params,
            **{k: v for d in self.inherited_params for k, v in d[1].items()},
        }

    def get_target(self) -> str:
        """
        Method returns whatever host has defined as target for connection.
        If name is just alias, and host has defined "hostname" then hostname will be returned
        When there is no hostname, only host name is returned
        """
        return self.name if "hostname" not in self.params else self.params["hostname"]

    def resolve_target(self) -> Tuple[str, bool]:
        """
        Method returns tuple of resolved IP address for this host, and error as bool value,
        that is set to true if host cannot be resolved
        """
        target = self.get_target()
        try:
            target_ip = socket.gethostbyname(target)
            return (target_ip, False)
        except socket.error:
            return ("", True)

    # Method for interaction with printing the object via Rich library
    # Each supported style should be located in defined folder (by default under "host_styles")
    # Each module must have "render" function, and return "rich renderable" object
    def __rich__(self):
        try:
            # If current host "print_style" is set to "panels", we will try
            # to import module from "./host_styles/panels.py"
            style = importlib.import_module(
                f"sshtmux.sshm.host_styles.{self.print_style}"
            )
            return style.render(self)
        except ModuleNotFoundError:
            return f"SSH_Host style [bright_red]'{self.print_style}'[/] is not implemented!"
        except Exception:
            if DEBUG_STYLES:
                console.print_exception(
                    show_locals=False, max_frames=1, suppress=[importlib]
                )
            return f"SSH_Host style [bright_red]'{self.print_style}'[/] is broken!"

    def __str__(self) -> str:
        return f"Group: {self.group} - Host: {self.name}"



import click

from sshtmux.sshm import (
    SSH_Config,
    SSH_Group,
    SSH_Host,
    complete_params,
    complete_ssh_group_names,
)
from sshtmux.sshm.sshutils import validate_ssh_params

# ------------------------------------------------------------------------------
# COMMAND: host create
# ------------------------------------------------------------------------------
SHORT_HELP = "Create new host"
LONG_HELP = """
Create new host and save it to config file

Command can be used to create a single HOST definition, and also set definitions within single command.
Later definitions can be changes with "sshm host set" commands.

If autocomplete is enabled, command will try to give suggestions for your inputs on definition of GROUP and well known PARAM names.
"""

# Parameters help:
INFO_HELP = "Set host info, can be set multiple times, or set to empty value to clear it (example: -i '')"
PARAM_HELP = "Sets parameter for the host, takes 2 values (<sshparam> <value>)."
GROUP_HELP = "Defined in which group host will be created, if not specified, 'default' group will be used"
FORCE_HELP = "Allows during host creation, to create group for host if target group is missing/not yet defined."
# ------------------------------------------------------------------------------


@click.command(name="create", short_help=SHORT_HELP, help=LONG_HELP)
@click.option("-i", "--info", multiple=True, help=INFO_HELP)
@click.option(
    "-p",
    "--parameter",
    nargs=2,
    multiple=True,
    help=PARAM_HELP,
    shell_complete=complete_params,
)
@click.option(
    "-g",
    "--group",
    "target_group_name",
    help=GROUP_HELP,
    shell_complete=complete_ssh_group_names,
)
@click.option("-f", "--force", is_flag=True, help=FORCE_HELP)
@click.argument("name")
@click.pass_context
def cmd(ctx, name, info, parameter, target_group_name, force):
    config: SSH_Config = ctx.obj

    if not target_group_name:
        if name == SSH_Config.GLOBAL_PATTERN_HOST_NAME:
            target_group_name = SSH_Config.GLOBAL_PATTERN_GROUP_NAME
        else:
            target_group_name = SSH_Config.DEFAULT_GROUP_NAME
    else:
        if not name.startswith(f"{target_group_name}-"):
            name = f"{target_group_name}-{name}"

    if config.check_host_by_name(name):
        print(f"Cannot create host '{name}' as it already exists in configuration!")
        ctx.exit(1)

    # Find group by name where to store config
    target_group_exists = config.check_group_by_name(target_group_name)

    if not target_group_exists and not force:
        print(
            f"Cannot create host '{name}' in group '{target_group_name}' since the group does not exist"
        )
        print("Create group first, or use '--force' option to create it automatically!")
        ctx.exit(1)
        # unreachable, but avoids issues with static checks
        exit(1)
    elif not target_group_exists:
        target_group = SSH_Group(name=target_group_name)
        config.groups.append(target_group)

        # Create group parttern host
        if name != SSH_Config.GLOBAL_PATTERN_HOST_NAME:
            new_pattern_host = SSH_Host(
                name=f"{target_group_name}-*",
                group=target_group_name,
                type="pattern",
                info=[],
            )
            target_group.patterns.append(new_pattern_host)
    else:
        target_group = config.get_group_by_name(target_group_name)

    # This is patter host
    target_type = "pattern" if "*" in name else "normal"
    new_host = SSH_Host(
        name=name, group=target_group_name, type=target_type, info=list(info)
    )

    # Add all passed parameters to config
    valid_ssh_params = validate_ssh_params(parameter)
    for param, value in valid_ssh_params.items():
        new_host.params[param] = value

    # Append new host to the group
    if new_host.type == "normal":
        target_group.hosts.append(new_host)
    else:
        target_group.patterns.append(new_host)

    # Generate new config
    config.generate_ssh_config().write_out()

    if not config.stdout:
        print(f"Created host: {name}")



import click
from rich import box
from rich.console import Console
from rich.table import Table

from sshtmux.sshm import SSH_Config

# ------------------------------------------------------------------------------
# COMMAND: group list
# ------------------------------------------------------------------------------
SHORT_HELP = "Lists all groups"
LONG_HELP = SHORT_HELP

# Parameters help:
# ------------------------------------------------------------------------------


@click.command(name="list", short_help=SHORT_HELP, help=LONG_HELP)
@click.pass_context
def cmd(ctx):
    config: SSH_Config = ctx.obj

    table = Table(box=box.SQUARE, style="grey35")
    table.add_column("Name", style="white")
    table.add_column("#Hosts", justify="right", style="bright_yellow")
    table.add_column("#Patterns", justify="right", style="bright_cyan")
    table.add_column("Desc", style="grey50")

    for group in config.groups_sorted:
        table.add_row(
            group.name, str(len(group.hosts)), str(len(group.patterns)), group.desc
        )

    console = Console()
    console.print(table)



import click

from .group import (
    group_create,
    group_delete,
    group_list,
    group_rename,
    group_set,
    group_show,
)


# ------------------------------------------------------------------------------
# GROUP Commands
# ------------------------------------------------------------------------------
@click.group(name="group", help="Command group for managing groups")
def ssh_group():
    pass


# // Linking other sub-commands
ssh_group.add_command(group_create.cmd)
ssh_group.add_command(group_delete.cmd)
ssh_group.add_command(group_list.cmd)
ssh_group.add_command(group_set.cmd)
ssh_group.add_command(group_show.cmd)
ssh_group.add_command(group_rename.cmd)



import click

from sshtmux.services.identities import KeyManager

HELP = "Generate a Fernet key - 32 url-safe"


@click.command(name="generate-key", short_help=HELP, help=HELP)
def cmd():
    key = KeyManager.generate_key().decode()
    output = f"""
    Be careful, when saving identities using a key, you cannot recover passwords without the key.
    If you have saved passwords and are using the ~/.config/sshtmux/identity.key file, you can configure the key of this file in the env var 'SSHTMUX_IDENTITY_KEY'.
    Otherwise you can use a new key belowS

    Set the env var SSHTMUX_IDENTITY_KEY

    Your key:

    {key}
    """
    print(output)



from typing import List

import click
from rich import box
from rich.console import Console
from rich.table import Table

from sshtmux.sshm import SSH_Config, SSH_Host

console = Console()

# ------------------------------------------------------------------------------
# COMMAND: host-list
# ------------------------------------------------------------------------------
SHORT_HELP = "List configured hosts"
LONG_HELP = """
List hosts from current configuration

Command allows displaying hosts with various properties, and filtering output list for easier searching.
Will display also usually not visible "INHERITANCE" of parameters trough "PATTERNS", and from
which pattern parameter for given host would be applied (inherited).

Normal output will try to show only most important information about listed hosts, while if requested,
VERBOSE option will try to show ALL defined/inherited parameters for all listed hosts.
Alternatively for details on only single host, use "sshm host show" command.

Host list can be limited with filters, to only show specific hosts that match NAME regex or GROUP regex.
When both NAME and GROUP regex are defined, output must satisfy both filters.
"""

# Parameters help:
GROUP_HELP = "Filter for host groups (regex)"
NAME_HELP = "Filter for host names (regex)"
VERBOSE_HELP = "Show verbose info (all parameters)"
# ------------------------------------------------------------------------------


@click.command(name="list", short_help=SHORT_HELP, help=LONG_HELP)
@click.option("-g", "--group", "group_filter", help=GROUP_HELP)
@click.option("-n", "--name", "name_filter", help=NAME_HELP)
@click.option("-v", "--verbose", is_flag=True, help=VERBOSE_HELP)
@click.pass_context
def cmd(ctx, group_filter, name_filter, verbose):
    config: SSH_Config = ctx.obj

    # Filter out groups and hosts if filters are defined via CLI
    filtered_groups = config.filter_config(group_filter, name_filter)

    if not filtered_groups:
        print("No host is matching any given filter!")
        ctx.exit(1)

    # This lists define host properties and which parameters will be displayed
    host_props = ["name", "group", "type"]
    params = ["hostname", "user"]

    # If output is verbose, we need to find all parameters, and add them to params list
    if verbose:
        flat_config: List[SSH_Host] = []
        for group in filtered_groups:
            for h in group.hosts + group.patterns:
                flat_config.append(h)
        for host in flat_config:
            for i_params in host.params:
                if i_params not in params:
                    params.append(i_params)

    header = host_props + ([f"param:{p}" for p in params])
    table = Table(*header, box=box.SQUARE, style="gray35")

    # Start adding rows
    for group in filtered_groups:
        # Iterate trough hosts and patters
        for host in group.hosts + group.patterns:
            host_params = []
            # Go trough list of all params we know are available across current host list
            for table_param in params:
                if table_param in host.params:
                    # Handle direct params, and handle if its a list or string
                    if isinstance(host.params[table_param], list):
                        host_params.append("\n".join(host.params[table_param]))
                    else:
                        host_params.append(host.params[table_param])
                else:
                    # Handle inherited params (only valid for "normal" hosts)
                    if host.inherited_params:
                        for pattern, i_params in host.inherited_params:
                            if table_param in i_params:
                                if isinstance(i_params[table_param], list):
                                    table_param = "\n".join(
                                        [
                                            f"{val}  ({pattern})"
                                            for val in i_params[table_param]
                                        ]
                                    )
                                    host_params.append(table_param)
                                else:
                                    host_params.append(
                                        f"[yellow]{i_params[table_param]}  ({pattern})[/]"
                                    )
                            else:
                                host_params.append("")
                    else:
                        host_params.append("")
            row = [host.__dict__[prop] for prop in host_props] + host_params
            table.add_row(*row) if host.type == "normal" else table.add_row(
                *row, style="cyan"
            )

    console.print(table)



import click

from sshtmux.sshm import SSH_Config, complete_ssh_host_names

# ------------------------------------------------------------------------------
# COMMAND: host rename
# ------------------------------------------------------------------------------
SHORT_HELP = "Rename existing host"
LONG_HELP = "Rename existing host in configuration"

# ------------------------------------------------------------------------------


@click.command(name="rename", short_help=SHORT_HELP, help=LONG_HELP)
@click.argument("name", shell_complete=complete_ssh_host_names)
@click.argument("new_name")
@click.pass_context
def cmd(ctx, name, new_name):
    config: SSH_Config = ctx.obj

    if not config.check_host_by_name(name):
        print(f"Cannot rename host '{name}' as it is not defined in configuration!")
        ctx.exit(1)
    if config.check_host_by_name(new_name):
        print(
            f"Cannot rename host '{name}' to '{new_name}' as new name is already used!"
        )
        ctx.exit(1)

    config.get_host_by_name(name)[0].name = new_name
    config.generate_ssh_config().write_out()

    if not config.stdout:
        print(f"Renamed host: {name} -> {new_name}")



import os
import re
from typing import Literal, Optional

from pydantic import BaseModel, IPvAnyAddress, field_validator, model_validator

CIPHERS = [
    "3des-cbc",
    "aes128-cbc",
    "aes192-cbc",
    "aes256-cbc",
    "aes128-ctr",
    "aes192-ctr",
    "aes256-ctr",
    "aes128-gcm@openssh.com",
    "aes256-gcm@openssh.com",
    "chacha20-poly1305@openssh.com",
]
CASIGNATUREALGORITHMS = [
    "+",
    "-",
    "ssh-ed25519",
    "ecdsa-sha2-nistp256",
    "ecdsa-sha2-nistp384",
    "ecdsa-sha2-nistp521",
    "sk-ssh-ed25519@openssh.com",
    "sk-ecdsa-sha2-nistp256@openssh.com",
    "rsa-sha2-512",
    "rsa-sha2-256",
]
KEX_ALGORITHMS = [
    "sntrup761x25519-sha512@openssh.com",
    "curve25519-sha256",
    "curve25519-sha256@libssh.org",
    "ecdh-sha2-nistp256",
    "ecdh-sha2-nistp384",
    "ecdh-sha2-nistp521",
    "diffie-hellman-group-exchange-sha256",
    "diffie-hellman-group16-sha512",
    "diffie-hellman-group18-sha512",
    "diffie-hellman-group14-sha256",
]
MAC_ALGORITHMS = [
    "hmac-sha2-256",
    "hmac-sha2-512",
    "hmac-sha1",
    "umac-64@openssh.com",
    "umac-128@openssh.com",
]
BASEALGORITHMS = [
    "ssh-ed25519-cert-v01@openssh.com",
    "ecdsa-sha2-nistp256-cert-v01@openssh.com",
    "ecdsa-sha2-nistp384-cert-v01@openssh.com",
    "ecdsa-sha2-nistp521-cert-v01@openssh.com",
    "sk-ssh-ed25519-cert-v01@openssh.com",
    "sk-ecdsa-sha2-nistp256-cert-v01@openssh.com",
    "rsa-sha2-512-cert-v01@openssh.com",
    "rsa-sha2-256-cert-v01@openssh.com",
    "ssh-ed25519",
    "ecdsa-sha2-nistp256",
    "ecdsa-sha2-nistp384",
    "ecdsa-sha2-nistp521",
    "sk-ssh-ed25519@openssh.com",
    "sk-ecdsa-sha2-nistp256@openssh.com",
    "rsa-sha2-512",
    "rsa-sha2-256",
]
IPQOS = [
    "af11",
    "af12",
    "af13",
    "af21",
    "af22",
    "af23",
    "af31",
    "af32",
    "af33",
    "af41",
    "af42",
    "af43",
    "cs0",
    "cs1",
    "cs2",
    "cs3",
    "cs4",
    "cs5",
    "cs6",
    "cs7",
    "ef",
    "le",
    "lowdelay",
    "throughput",
    "reliability",
    "none",
]
KBDINTERACTIVEDEVICES = ["pam", "bsdauth", "skey", "null", "default"]
LOG_LEVEL = [
    "QUIET",
    "FATAL",
    "ERROR",
    "INFO",
    "VERBOSE",
    "DEBUG",
    "DEBUG1",
    "DEBUG2",
    "DEBUG3",
]
PREFERREDAUTHENTICATIONS = [
    "gssapi-with-mic",
    "hostbased",
    "publickey",
    "keyboard-interactive",
    "password",
]
STRICTHOSTKEYCHECKING = Literal["yes", "no", "ask", "accept-new", "off"]
YES_NO = Literal["yes", "no"]
SYSLOGFACILITY = Literal[
    "DAEMON",
    "USER",
    "AUTH",
    "LOCAL0",
    "LOCAL1",
    "LOCAL2",
    "LOCAL3",
    "LOCAL4",
    "LOCAL5",
    "LOCAL6",
    "LOCAL7",
]
PORT_RANGE = (1, 65535)
SSH_HOSTNAME_REGEX = r"^(?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)$|^([a-zA-Z0-9]|[a-zA-Z0-9][a-zA-Z0-9\-]*[a-zA-Z0-9])(\.[a-zA-Z0-9\-]+)*$|^([0-9a-fA-F]{1,4}:){7}([0-9a-fA-F]{1,4})$"


class SSHParams(BaseModel):
    AddKeysToAgent: Optional[Literal["yes", "no", "ask", "confirm"]] = None
    AddressFamily: Optional[Literal["any", "inet", "inet6"]] = None
    BatchMode: Optional[YES_NO] = None
    BindAddress: Optional[IPvAnyAddress] = None
    BindInterface: Optional[str] = None
    CanonicalDomains: Optional[str] = None
    CanonicalizeFallbackLocal: Optional[YES_NO] = None
    CanonicalizeHostname: Optional[YES_NO] = None
    CanonicalizeMaxDots: Optional[str] = None
    CanonicalizePermittedCNAMEs: Optional[str] = None
    CASignatureAlgorithms: Optional[str] = None
    CertificateFile: Optional[str] = None
    ChannelTimeout: Optional[str] = None
    CheckHostIP: Optional[YES_NO] = None
    Ciphers: Optional[str] = None
    ClearAllForwardings: Optional[YES_NO] = None
    Compression: Optional[YES_NO] = None
    ConnectionAttempts: Optional[str] = None
    ConnectTimeout: Optional[str] = None
    ControlMaster: Optional[Literal["yes", "no", "ask", "auto", "autoask"]] = None
    ControlPath: Optional[str] = None
    ControlPersist: Optional[YES_NO] = None
    DynamicForward: Optional[str] = None
    EnableEscapeCommandline: Optional[str] = None
    EnableSSHKeysign: Optional[YES_NO] = None
    EscapeChar: Optional[str] = None
    ExitOnForwardFailure: Optional[YES_NO] = None
    FingerprintHash: Optional[Literal["sha256", "md5"]] = None
    ForkAfterAuthentication: Optional[YES_NO] = None
    ForwardAgent: Optional[YES_NO] = None
    ForwardX11: Optional[YES_NO] = None
    ForwardX11Timeout: Optional[str] = None
    ForwardX11Trusted: Optional[YES_NO] = None
    GatewayPorts: Optional[YES_NO] = None
    GlobalKnownHostsFile: Optional[str] = None
    GSSAPIAuthentication: Optional[YES_NO] = None
    GSSAPIKeyExchange: Optional[str] = None
    GSSAPIClientIdentity: Optional[str] = None
    GSSAPIDelegateCredentials: Optional[str] = None
    GSSAPIKexAlgorithms: Optional[str] = None
    GSSAPIRenewalForcesRekey: Optional[str] = None
    GSSAPIServerIdentity: Optional[str] = None
    GSSAPITrustDns: Optional[str] = None
    HashKnownHosts: Optional[YES_NO] = None
    HostbasedAcceptedAlgorithms: Optional[str] = None
    HostbasedAuthentication: Optional[YES_NO] = None
    HostKeyAlgorithms: Optional[str] = None
    HostKeyAlias: Optional[str] = None
    Hostname: Optional[str] = None
    IdentitiesOnly: Optional[YES_NO] = None
    IdentityAgent: Optional[str] = None
    IdentityFile: Optional[str] = None
    IgnoreUnknown: Optional[str] = None
    Include: Optional[str] = None
    IPQoS: Optional[str] = None
    KbdInteractiveAuthentication: Optional[YES_NO] = None
    KbdInteractiveDevices: Optional[str] = None
    KexAlgorithms: Optional[str] = None
    KnownHostsCommand: Optional[str] = None
    LocalCommand: Optional[str] = None
    LocalForward: Optional[str] = None
    LogLevel: Optional[str] = None
    LogVerbose: Optional[str] = None
    MACs: Optional[str] = None
    NoHostAuthenticationForLocalhost: Optional[YES_NO] = None
    NumberOfPasswordPrompts: Optional[str] = None
    ObscureKeystrokeTiming: Optional[YES_NO] = None
    PasswordAuthentication: Optional[YES_NO] = None
    PermitLocalCommand: Optional[YES_NO] = None
    PermitRemoteOpen: Optional[str] = None
    PKCS11Provider: Optional[str] = None
    Port: Optional[str] = None
    PreferredAuthentications: Optional[str] = None
    ProxyCommand: Optional[str] = None
    ProxyJump: Optional[str] = None
    ProxyUseFdpass: Optional[YES_NO] = None
    PubkeyAcceptedAlgorithms: Optional[str] = None
    PubkeyAuthentication: Optional[Literal["yes", "no", "unbound", "host-bound"]] = None
    RekeyLimit: Optional[str] = None
    RemoteCommand: Optional[str] = None
    RemoteForward: Optional[str] = None
    RequestTTY: Optional[Literal["yes", "no", "force", "auto"]] = None
    RequiredRSASize: Optional[str] = None
    RevokedKeys: Optional[str] = None
    RevokedHostKeys: Optional[str] = None
    SecurityKeyProvider: Optional[str] = None
    SendEnv: Optional[str] = None
    ServerAliveCountMax: Optional[str] = None
    ServerAliveInterval: Optional[str] = None
    SessionType: Optional[str] = None
    StdinNull: Optional[YES_NO] = None
    StreamLocalBindMask: Optional[str] = None
    StreamLocalBindUnlink: Optional[YES_NO] = None
    StrictHostKeyChecking: Optional[STRICTHOSTKEYCHECKING] = None
    SyslogFacility: Optional[SYSLOGFACILITY] = None
    Tag: Optional[str] = None
    TCPKeepAlive: Optional[YES_NO] = None
    Tunnel: Optional[Literal["yes", "no", "point-to-point", "ethernet"]] = None
    TunnelDevice: Optional[str] = None
    UpdateHostKeys: Optional[Literal["yes", "no", "ask"]] = None
    User: Optional[str] = None
    UserKnownHostsFile: Optional[str] = None
    VerifyHostKeyDNS: Optional[STRICTHOSTKEYCHECKING] = None
    VisualHostKey: Optional[YES_NO] = None
    XAuthLocation: Optional[str] = None

    class Config:
        extra = "forbid"

    @staticmethod
    def validate_list_values(v, list_values, divisor=","):
        if v:
            values = v.split(divisor)
            for value in values:
                if value not in list_values:
                    raise ValueError(f"Value list should be one of: {list_values}")
        return v

    @staticmethod
    def is_valid_time_format(value: str) -> None:
        if value == "0":
            return
        pattern = re.compile(r"^(\d+[smhdwSMHDW]?)+$")
        is_valid = bool(pattern.match(value))
        if not is_valid:
            raise ValueError(f"Invalid Time Format: {value}")

    @staticmethod
    def is_valid_bind(value: str) -> bool:
        pattern = r"^((\d{1,3}\.){3}\d{1,3}|\[::1\]|0\.0\.0\.0)?(:)?(\d+)$"
        match = re.match(pattern, value)
        if not match:
            raise ValueError("Invalid format. must be [bind_address:]port")

        port_str = match.group(4)
        if port_str is None:
            raise ValueError(f"Port not found: {value}")

        port = int(port_str)
        if not (PORT_RANGE[0] <= port <= PORT_RANGE[1]):
            raise ValueError(f"Port must be (1-65535): {port}")

    @staticmethod
    def is_valid_string_integer(value: str) -> None:
        if not isinstance(value, str) or value.startswith("0"):
            raise ValueError(f"Invalid number. Can not starts with 0: {value}")
        try:
            value = int(value)
        except ValueError:
            raise ValueError(f"Should be a integer number: {value}")
        if not value >= 1:
            raise ValueError("Shoud be greater then 0")

    @model_validator(mode="before")
    def validate_all_fields(cls, values):
        new_values = {k.strip(): v.strip() for k, v in values.items()}
        for field, value in new_values.items():
            if value == "":
                raise ValueError(f"{field} Can not be empty string")
        return new_values

    @field_validator("Hostname")
    def validate_hostname(cls, v):
        try:
            IPvAnyAddress(v)
        except ValueError:
            domain_pattern = re.compile(SSH_HOSTNAME_REGEX)
            if not domain_pattern.match(v):
                raise ValueError(f"Hostname invalid: {v}")
        return v

    @field_validator("Port")
    def validate_port(cls, v):
        SSHParams.is_valid_string_integer(v)
        if not (PORT_RANGE[0] <= int(v) <= PORT_RANGE[1]):
            raise ValueError(f"Port must be (1-65535): {v}")
        return v

    @field_validator("ControlPersist")
    def validate_controlpersist(cls, v):
        if v in ["yes", "no", "auto"]:
            return v
        SSHParams.is_valid_time_format(v)
        return v

    @field_validator("DynamicForward", "PermitRemoteOpen")
    def validate_dynamic_forward(cls, v):
        SSHParams.is_valid_bind(v)

    @field_validator(
        "ConnectionAttempts",
        "ConnectTimeout",
        "ServerAliveCountMax",
        "ServerAliveInterval",
        "CanonicalizeMaxDots",
        "NumberOfPasswordPrompts",
        "RequiredRSASize",
    )
    def validate_greater_then_0(cls, v):
        SSHParams.is_valid_string_integer(v)
        return v

    @field_validator("ChannelTimeout", "ForwardX11Timeout")
    def validate_greater_then_0_and_time_formats(cls, v):
        SSHParams.is_valid_string_integer(v)
        SSHParams.is_valid_time_format(v)
        return v

    @field_validator("CASignatureAlgorithms")
    def validate_casignaturealgorithms(cls, v):
        SSHParams.validate_list_values(v, CASIGNATUREALGORITHMS)
        return v

    @field_validator(
        "HostbasedAuthentication",
        "HostbasedAcceptedAlgorithms",
        "HostKeyAlgorithms",
        "PubkeyAcceptedAlgorithms",
    )
    def validate_hostbasedacceptedalgorithms(cls, v):
        SSHParams.validate_list_values(v, BASEALGORITHMS)
        return v

    @field_validator("Ciphers")
    def validate_ciphers(cls, v):
        SSHParams.validate_list_values(v, CIPHERS)
        return v

    @field_validator("KexAlgorithms")
    def validate_kexs(cls, v):
        SSHParams.validate_list_values(v, KEX_ALGORITHMS)
        return v

    @field_validator("IPQoS")
    def validate_ipqos(cls, v):
        SSHParams.validate_list_values(v, IPQOS, " ")
        return v

    @field_validator("KbdInteractiveDevices")
    def validate_kbdinteractivedevices(cls, v):
        SSHParams.validate_list_values(v, KBDINTERACTIVEDEVICES)
        return v

    @field_validator("LogLevel")
    def validate_loglevel(cls, v):
        SSHParams.validate_list_values(v, LOG_LEVEL)
        return v

    @field_validator("MACs")
    def validate_macs(cls, v):
        SSHParams.validate_list_values(v, MAC_ALGORITHMS)
        return v

    @field_validator("PreferredAuthentications")
    def validate_preferredauthentications(cls, v):
        SSHParams.validate_list_values(v, PREFERREDAUTHENTICATIONS)
        return v

    @field_validator("RekeyLimit")
    def validate_rekey_limit(cls, v):
        pattern = r"^(\d+)([kKmMgGhHs])$"
        match = re.match(pattern, v.strip())
        if match:
            number = int(match.group(1))
            unit = match.group(2).lower()
            valid_units = ["k", "m", "g", "h", "s"]
            if unit in valid_units and number > 0:
                return True
        raise ValueError(f"Invalid Format: {v}")

    @field_validator(
        "IdentityFile",
        "CertificateFile",
        "UserKnownHostsFile",
        "GlobalKnownHostsFile",
        "PKCS11Provider",
        "ControlPath",
        "RevokedHostKeys",
        "RevokedKeys",
        "SecurityKeyProvider",
    )
    def validate_path(cls, v):
        if v and not os.path.exists(v):
            raise ValueError(f"File or Path does not exits: {v}")
        return v

    @field_validator("LocalForward", "RemoteForward")
    def validate_forward(cls, v):
        pattern = r"^(\d+)\s+((localhost|[\w.-]+|\[.*\])):(\d+)$"
        match = re.match(pattern, v)

        if not match:
            raise ValueError(
                "Invalid format. Expected format: <port> <host>:<port> (e.g., '4000 localhost:4000')"
            )

        local_port, host, _, remote_port = match.groups()
        if local_port.startswith("0"):
            raise ValueError(f"Invalid Port number: {local_port}")
        if remote_port.startswith("0"):
            raise ValueError(f"Invalid Port number: {remote_port}")

        local_port = int(local_port)
        remote_port = int(remote_port)

        if not (PORT_RANGE[0] <= local_port <= PORT_RANGE[1]):
            raise ValueError(
                f"Local port must be between {PORT_RANGE[0]} and {PORT_RANGE[1]}: {local_port}"
            )
        if not (PORT_RANGE[0] <= remote_port <= PORT_RANGE[1]):
            raise ValueError(
                f"Remote port must be between {PORT_RANGE[0]} and {PORT_RANGE[1]}: {remote_port}"
            )
        if not re.match(SSH_HOSTNAME_REGEX, host):
            raise ValueError(
                f"Invalid hostname format for SSH: {host}. It must follow the SSH hostname rules."
            )
        return v



from rich import box
from rich.panel import Panel
from rich.table import Table

from ..ssh_host import SSH_Host


# ------------------------------------------------------------------------------
# Render host data in nested table with separate SSH parameters
# ------------------------------------------------------------------------------
def render(host: SSH_Host):
    out_type = host.type if host.type == "normal" else f"[cyan]{host.type}[/]"
    out_info = "\n".join(host.info) if host.info else "- No info defined - "

    outer_table = Table(box=box.SQUARE, style="grey35", show_header=False)
    outer_table.add_column("Parameter")
    outer_table.add_column("Value")

    outer_table.add_row("Name", host.name)
    outer_table.add_row("Group", host.group)
    outer_table.add_row("Type", out_type)
    outer_table.add_row("Info", Panel(out_info, border_style="grey35", style="grey50"))

    param_table = Table(
        box=box.SQUARE, style="grey35", show_header=True, show_edge=True, expand=True
    )
    param_table.add_column("Param")
    param_table.add_column("Value")
    param_table.add_column("Inherited-from")

    # Add rows for SSH Config parameters
    for key, value in host.params.items():
        output_value = value if not isinstance(value, list) else "\n".join(value)
        param_table.add_row(key, output_value)

    # Add rows for inherited SSH Config parameters
    for pattern, pattern_params in host.inherited_params:
        for param, value in pattern_params.items():
            if param not in host.params:
                output_value = (
                    value if not isinstance(value, list) else "\n".join(value)
                )
                param_table.add_row(param, output_value, pattern, style="yellow")

    outer_table.add_row("SSH Params", param_table)
    return outer_table



import click
from rich import box
from rich.console import Console
from rich.table import Table

from sshtmux.exceptions import IdentityException
from sshtmux.services.identities import PasswordManager

HELP = "List all Identities"


@click.command(name="list", short_help=HELP, help=HELP)
def cmd():
    password_manager = PasswordManager()
    identies = password_manager.get_identities()
    try:
        identies = password_manager.get_identities()
    except IdentityException as e:
        print(str(e))
        return
    console = Console()
    table = Table(box=box.SQUARE, style="gray35")
    table.add_column("Identity")
    for identity in identies:
        table.add_row(identity)

    console.print(table)



import click
from rich.console import Console

from sshtmux.core.config import T_Host_Style, settings
from sshtmux.sshm import (
    SSH_Config,
    complete_ssh_host_names,
    complete_styles,
    generate_graph,
    trace_jumphosts,
)

console = Console()
styles_str = ",".join(T_Host_Style.__args__)

# ------------------------------------------------------------------------------
# COMMAND: host show
# ------------------------------------------------------------------------------
SHORT_HELP = "Show current host configuration"
LONG_HELP = """
Shows details about host and its configuration

Command can generate nice representation of configuration for given HOST.
If HOST is using "proxyjump" properties, command will try to collect also all info
from intermediate hosts/proxies as well.

Additionally when using --graph option, command can "draw" visualization of
connection path, and defined end-to-end tunnels
"""

# Parameters help:
FOLLOW_HELP = "Follow and displays all connected hosts via proxyjump (works only for locally defined hosts)"
GRAPH_HELP = "Shows connection to target as graph with tunnels visualizations"
STYLE_HELP = f"Select output rendering style for host details: ({styles_str}), (default: {settings.sshtmux.SSHTMUX_HOST_STYLE})"
# ------------------------------------------------------------------------------


@click.command(name="show", short_help=SHORT_HELP, help=LONG_HELP)
@click.option(
    "-s",
    "--style",
    default="",
    help=STYLE_HELP,
    shell_complete=complete_styles,
)
@click.option("-g", "--graph", is_flag=True, help=GRAPH_HELP)
@click.argument("name", shell_complete=complete_ssh_host_names)
@click.pass_context
def cmd(ctx: click.core.Context, name: str, style: str, graph: bool):
    config: SSH_Config = ctx.obj

    # Define host print style from CLI or config or default
    if not style:
        if "host-style" in config.opts:
            style = config.opts["host-style"]
        else:
            style = settings.sshtmux.SSHTMUX_HOST_STYLE

    if not config.check_host_by_name(name):
        print(
            f"Cannot get info for host '{name}' as it is not defined in configuration!"
        )
        ctx.exit(1)

    # Get info about linked hosts (required for graph and follow)
    traced_hosts = trace_jumphosts(name, config, ctx, style)

    # Normal show, no linked graphs needed
    console.print(traced_hosts[0])

    if graph:
        console.print("\n")
        console.print(generate_graph(traced_hosts), "")



import click
from libtmux.exc import LibTmuxException

from sshtmux.services.tmux import Tmux

HELP = "Run commands on Tmux (Executed only by Tmux)"


@click.command(name="run", short_help=HELP, help=HELP)
@click.argument("session-name")
@click.argument("window-index")
@click.argument("panel-index")
@click.argument("cmd-ref")
def cmd(session_name, window_index, panel_index, cmd_ref):
    tmux = Tmux()
    try:
        tmux.execute_host_cmd(session_name, window_index, panel_index, cmd_ref)
    except LibTmuxException as e:
        print(str(e))



from rich import box
from rich.table import Table

from ..ssh_host import SSH_Host


# ------------------------------------------------------------------------------
# Render host data in single flat 3-column table
# ------------------------------------------------------------------------------
def render(host: SSH_Host):
    out_type = host.type if host.type == "normal" else f"[cyan]{host.type}[/]"
    out_info = "\n".join(host.info) if host.info else "- No info defined - "

    table = Table(box=box.SQUARE, style="grey39")
    table.add_column("Parameter")
    table.add_column("Value")
    table.add_column("Inherited-from")

    table.add_row("Name", host.name)
    table.add_row("Group", host.group)
    table.add_row("Type", out_type)
    table.add_row("Info", out_info, style="grey50")

    # Add rows for SSH Config parameters
    for key, value in host.params.items():
        output_value = value if not isinstance(value, list) else "\n".join(value)
        table.add_row(f"Param:{key}", output_value)

    # Add rows for inherited SSH Config parameters
    for pattern, pattern_params in host.inherited_params:
        for param, value in pattern_params.items():
            if not param in host.params:
                output_value = (
                    value if not isinstance(value, list) else "\n".join(value)
                )
                table.add_row(f"param:{param}", output_value, pattern, style="yellow")

    return table



import click
from libtmux.exc import LibTmuxException
from rich import box
from rich.console import Console
from rich.table import Table

from sshtmux.services.snippets import get_snippets_files

HELP = "List Snippets"


@click.command(name="list", short_help=HELP, help=HELP)
def cmd():
    console = Console()
    table = Table(box=box.SQUARE, style="gray35")
    table.add_column("Snippets")
    snippets, _ = get_snippets_files()
    for snippet in snippets:
        if snippet == "Cancel":
            continue
        table.add_row(snippet)
    console.print(table)



from dataclasses import dataclass, field
from typing import List

from .ssh_host import SSH_Host


@dataclass
class SSH_Group:
    """Class for SSH Group config structure"""

    name: str
    desc: str = ""
    info: list = field(default_factory=list)
    hosts: List[SSH_Host] = field(default_factory=list)
    patterns: List[SSH_Host] = field(default_factory=list)

    print_style: str = ""

    def __rich__(self):
        pass



import copy
import fnmatch
import logging
import os
import re
from pathlib import Path
from typing import List, Optional, Tuple

from ..core.config import settings
from .ssh_group import SSH_Group
from .ssh_host import SSH_Host

logging.basicConfig(level=logging.INFO)


class SSH_Config:
    """
    SSH Configuration class

    Main class for handling SSH configuration, reading from file, parsing and
    generating and writing contents back to SSH configuration file
    """

    DEFAULT_GROUP_NAME: str = "default"
    GLOBAL_PATTERN_GROUP_NAME: str = "global_pattern"
    GLOBAL_PATTERN_HOST_NAME: str = "*"

    def __init__(self, config_lines: List[str] = [], stdout: bool = False):
        self.ssh_config_file: str = os.path.expanduser(settings.ssh.SSH_CONFIG_FILE)
        self.ssh_config_lines: List[str] = config_lines

        # configuration representation (array of SSH groups?)
        self.groups: List[SSH_Group] = [
            SSH_Group(name=self.DEFAULT_GROUP_NAME, desc="Default group")
        ]
        self.global_pattern_group = SSH_Group(
            name=self.GLOBAL_PATTERN_GROUP_NAME, desc="Global pattern"
        )

        # options
        self.stdout: bool = stdout
        self.opts: dict = {}

        # parsing "cache" info
        self.current_grindex: int = 0
        self.current_group: str = self.DEFAULT_GROUP_NAME
        self.current_host: Optional[SSH_Host] = None
        self.current_host_info: list = []

    @property
    def groups_sorted(self):
        if self.groups[-1].name == self.GLOBAL_PATTERN_GROUP_NAME:
            self.groups.insert(0, self.groups.pop(-1))
        print(self.groups[0])
        return self.groups

    def read(self):
        """
        Read content of SSH config file
        """
        if not Path(self.ssh_config_file).exists():
            print(f"File {self.ssh_config_file} not found! Creating...")
            Path(self.ssh_config_file).touch()

        with open(self.ssh_config_file, "r") as fh:
            self.ssh_config_lines = fh.readlines()
        return self

    def _config_flush_host(self) -> None:
        """
        Internal function used to flush host configuration while parsing config file
        """
        if self.current_host:
            if self.current_host.name == "*":
                if len(self.global_pattern_group.patterns) == 0:
                    self.current_host.group = self.global_pattern_group.name
                    self.global_pattern_group.patterns.append(self.current_host)
                return

            if self.current_host.type == "normal":
                self.groups[self.current_grindex].hosts.append(self.current_host)
            else:
                self.groups[self.current_grindex].patterns.append(self.current_host)
            # Reset "cache" since we flushed host info
            self.current_host = None

    def _sort_groups(self):
        base_groups_list = [self.DEFAULT_GROUP_NAME, self.GLOBAL_PATTERN_GROUP_NAME]
        base_groups = [g for g in self.groups if g.name in base_groups_list]
        other_groups = [g for g in self.groups if g.name not in base_groups_list]

        other_groups.sort(key=lambda g: g.name)

        self.groups = (
            [g for g in base_groups if g.name == self.DEFAULT_GROUP_NAME]
            + other_groups
            + [g for g in base_groups if g.name == self.GLOBAL_PATTERN_GROUP_NAME]
        )

    def parse(self):
        """
        Parse config lines one by one and generate configuration structure
        """
        # Parse each line of the configuration, line by line
        for config_line_index, line in enumerate(self.ssh_config_lines):
            line = line.strip()  # remove start and end whitespace

            # Skip empty lines, go to next...
            if not line:
                continue

            # Reworked parsing method... special "meta" keys not needed, regex now parses words
            # so its much more "freely" in reading from config file writing is still in legacy style
            # in future, output config write styles could be changed...
            if line.startswith("#"):
                match = re.search(
                    r"^#[\s@]*(group|desc|info|host|config)[\s:]+(.+)$", line
                )

                # In case special comment is unreadable
                if not match:
                    logging.debug(f"DROP COMMENT: '{line}'")
                    continue

                metadata, value = match.groups()

                if value == self.GLOBAL_PATTERN_GROUP_NAME:
                    continue

                if metadata == "config":
                    # Config options are configured as key=value within config line...
                    logging.debug(
                        f"META: Config line found '{value}' Caching for next host definition...'"
                    )
                    conf_key, conf_val = value.split("=")
                    self.opts[conf_key] = conf_val
                    continue

                elif metadata == "group":
                    # New group found... flush any previous data and create new baseline
                    self._config_flush_host()

                    logging.debug(f"META: Starting new group: {value}")
                    self.groups.append(SSH_Group(name=value))

                    self.current_grindex = len(self.groups) - 1
                    self.current_group = value
                    logging.debug(f"META: Group index set to: {self.current_grindex}")
                    continue

                elif metadata == "desc":
                    logging.debug(
                        f"META: Setting 'desc' param to '{value}' for group '{self.groups[self.current_grindex].name}'"
                    )
                    self.groups[self.current_grindex].desc = value
                    continue

                elif metadata == "info":
                    logging.debug(
                        f"META: Setting 'info' param to '{value}' for group '{self.groups[self.current_grindex].name}'"
                    )
                    self.groups[self.current_grindex].info.append(value)
                    continue

                elif metadata == "host":
                    logging.debug(
                        f"META: Host comment found '{value}' Caching for next host definition...'"
                    )
                    self.current_host_info.append(value)
                    continue

                else:
                    logging.warning(
                        f"META: Unhandled metadata '{metadata}' on SSH-config line number: {config_line_index}"
                    )
                    continue

            # Here we expect only normal ssh config lines "Host" is usually the keyword that begins definition
            # if we find any other keyword before first host keyword is defined, configuration is wrong probably
            match = re.search(r"^(\w+)\s+(.+)$", line)
            if not match:
                logging.warning(
                    f"KEYWORD: Incorrect configuration line '{line}' on SSH-config line number: {config_line_index}"
                )
                continue

            keyword, value = match.groups()

            # --- Found "host" keyword, that defines new block, usually followed with name
            if keyword.lower() == "host":
                self._config_flush_host()

                host_type = "pattern" if "*" in value else "normal"
                logging.debug(f"Host '{value}' is '{host_type}' type!")

                self.current_host = SSH_Host(
                    name=value,
                    group=self.current_group,
                    type=host_type,
                    info=self.current_host_info,
                )
                # Reset global host info cache when we find new host (from this line, any host comments will apply to next host)
                self.current_host_info = []
                continue
            else:
                # any other normal line we just use as it is, wrong or not... :)
                # Currently there is no support for keyword validation
                if not self.current_host:
                    logging.warning(
                        f"Config info without Host definition on SSH-config line number {config_line_index}"
                    )
                    exit(1)
                else:
                    logging.debug(
                        f"Config keyword for host '{self.current_host}': {keyword} -> {value}"
                    )
                    self.current_host.params[keyword] = value
                    continue

        # Last entries must be flushed manually as there are no new "hosts" to trigger storing parsed data into config struct
        self._config_flush_host()

        if len(self.global_pattern_group.patterns) > 0:
            self.groups.append(self.global_pattern_group)
        for group in self.groups:
            for host in group.hosts:
                inherited_params: List[Tuple[str, dict]] = []
                if host.type == "normal":
                    inherited_params = self.find_inherited_params(host.name)
                    host.inherited_params = inherited_params

        self._sort_groups()
        return self

    def generate_ssh_config(self):
        """
        SSH config generation function

        Function takes config data structure, and target file information.
        Then generates SSH config compatible file with all data, compatible with
        internal object model.
        """
        SSHCONFIG_INDENT = "    "
        SSHCONFIG_META_PREFIX = "@"
        SSHCONFIG_META_SEPARATOR = ": "

        # First we lines before we flush them into file
        lines: List[str] = ["#<<<<< SSH Config file managed by SSHTmux >>>>>\n"]

        # Dump any saved configuration
        for option in self.opts:
            lines.append(
                f"#{SSHCONFIG_META_PREFIX}config{SSHCONFIG_META_SEPARATOR}{option}={self.opts[option]}\n"
            )

        # Add separation from header/config and rest of ssh-config
        lines.append("\n")

        # Render all groups
        self._sort_groups()
        for group in self.groups:
            # Ship default group as it does not have to be specified
            render_header = False if group.name == self.DEFAULT_GROUP_NAME else True

            if render_header:
                # Add extra blank line when outputting new group header
                lines.append("\n")
                # Start header line for the group with known metadata
                lines.append(f"#{'-'*79}\n")  # add horizontal decoration line
                lines.append(
                    f"#{SSHCONFIG_META_PREFIX}group{SSHCONFIG_META_SEPARATOR}{group.name}\n"
                )
                if group.desc:
                    lines.append(
                        f"#{SSHCONFIG_META_PREFIX}desc{SSHCONFIG_META_SEPARATOR}{group.desc}\n"
                    )
                for info in group.info:
                    lines.append(
                        f"#{SSHCONFIG_META_PREFIX}info{SSHCONFIG_META_SEPARATOR}{info}\n"
                    )
                lines.append(f"#{'-'*79}\n")  # add horizontal decoration line

            # Append hosts and patterns items from group
            for host in group.hosts + group.patterns:
                # If there is host-info assigned to host, add it before adding "host" definition
                for host_info in host.info:
                    lines.append(
                        f"#{SSHCONFIG_META_PREFIX}host{SSHCONFIG_META_SEPARATOR}{host_info}\n"
                    )

                # Add "host" line definition
                lines.append(f"Host {host.name}\n")

                # Add all assigned host params
                for token, value in host.params.items():
                    if type(value) is str:
                        lines.append(f"{SSHCONFIG_INDENT}{token} {value}\n")
                    elif type(value) is list:
                        for v in value:
                            lines.append(f"{SSHCONFIG_INDENT}{token} {v}\n")
                    else:
                        raise Exception("Host parameter is not 'str' or 'list'!!!")

                # Add newline after host definition
                lines.append("\n")

        # Store output lines
        self.ssh_config_lines = lines
        return self

    def write_out(self) -> None:
        """
        Write generated SSH config to target file
        """
        if self.stdout:
            print("".join(self.ssh_config_lines))
        else:
            with open(self.ssh_config_file, "w") as out:
                out.writelines(self.ssh_config_lines)

    def check_group_by_name(self, name: str) -> bool:
        """
        Check if specific group name is present in configuration
        """
        is_valid_name, message = self.validate_name(name)
        if not is_valid_name:
            print(message)
            exit(1)

        for group in self.groups:
            if group.name == name:
                return True
        return False

    def get_group_by_name(self, name: str) -> SSH_Group:
        """
        Find group in configuration that matches the name (strict match, one only!)
        On success returns matched group, on fail depending on 'throw_on_fail' flag
        function will either return 'None' or will throw exception
        """
        for group in self.groups:
            if group.name == name:
                return group
        raise Exception(f"Requested group '{name}' not found in the SSH configuration")

    def check_host_by_name(self, name: str, validate_names: bool = True) -> bool:
        """
        Check if specific host name is present in configuration
        """
        if validate_names:
            is_valid_name, message = self.validate_name(name)
            if not is_valid_name:
                print(message)
                exit(1)

        for group in self.groups:
            all_hosts = group.hosts + group.patterns
            for host in all_hosts:
                if host.name == name:
                    return True
        return False

    def get_host_by_name(self, name: str) -> Tuple[SSH_Host, SSH_Group]:
        """
        Find host in configuration that matches the name (strict match, one only!)
        On success returns host and his assigned group, on fail depending on 'throw_on_fail' flag
        function will either return ('None','None') or will throw exception
        """
        for group in self.groups:
            all_hosts = group.hosts + group.patterns
            for host in all_hosts:
                if host.name == name:
                    return host, group
        raise Exception(f"Requested host '{name}' not found in the SSH configuration")

    def get_all_host_names(self) -> List[str]:
        """
        Return all host names from current configuration
        Useful for auto-completion, or for quick checking if name already exists
        """
        all_hosts: List[str] = []
        for group in self.groups:
            for host in group.hosts + group.patterns:
                all_hosts.append(host.name)
        return all_hosts

    def get_all_group_names(self) -> List[str]:
        """
        Return all group names from current configuration
        Useful for auto-completion, or for quick checking if name already exists
        """
        all_groups: List[str] = []
        for group in self.groups:
            all_groups.append(group.name)
        return all_groups

    def find_inherited_params(self, host_name: str) -> List[Tuple[str, dict]]:
        """
        Given a host name, finds and returns list of 2-item tuples, where first item is name of pattern from
        which params are inherited, and second item is parameters dictionary from the pattern
        """
        inherited: List[Tuple[str, dict]] = []
        for group in self.groups:
            for pattern in group.patterns:
                # Check if any one of pattern (from all groups) will match host name
                if fnmatch.fnmatch(host_name, pattern.name):
                    inherited.append((pattern.name, pattern.params))

        return inherited

    def filter_config(self, group_filter: str, name_filter: str) -> List[SSH_Group]:
        """
        Function takes optional group and name regex, and if they are not None or empty,
        function creates a new list of SSH groups and their SSH hosts, but with all
        non-matching items removed.
        """
        filtered_groups: List[SSH_Group] = []
        for group in self.groups:
            # If group filter is defined, check if current group matches the name to progress
            if group_filter:
                group_match = re.search(group_filter, group.name)
                if not group_match:
                    continue

            # When group is not skipped, check if name filter is used, and filter out groups
            if name_filter:
                # Make a new copy of group, so we dont mess original config
                group_copy = copy.copy(group)
                group_copy.hosts = []
                group_copy.patterns = []
                include_group = False

                for host in group.hosts + group.patterns:
                    match = re.search(name_filter, host.name)
                    if match:
                        include_group = True
                        if host.type == "normal":
                            group_copy.hosts.append(host)
                        else:
                            group_copy.patterns.append(host)
                if include_group:
                    filtered_groups.append(group_copy)
            else:
                filtered_groups.append(group)
        return filtered_groups

    def move_host_to_group(
        self, found_host: SSH_Host, found_group: SSH_Group, target_group: SSH_Group
    ) -> None:
        """
        Function that moves host from one group to other group
        """
        if found_host.type == "normal":
            target_group.hosts.append(found_host)
            found_group.hosts.remove(found_host)
        else:
            target_group.patterns.append(found_host)
            found_group.patterns.remove(found_host)

    def validate_name(self, name):
        if not settings.ssh.SSH_VALIDATE_SSHCONFIG:
            return True, None

        if not (3 <= len(name) <= 50):
            return False, "The name must be between 3 and 50 characters."

        if not re.match(r"^[a-zA-Z0-9_\-\*]+$", name):
            return (
                False,
                "The name cannot contain special characters",
            )
        return True, None



from pathlib import Path
from typing import Literal

from pydantic import BaseModel
from pydantic_settings import BaseSettings

USER_DIR = Path.home()
SSHTMUX_BASEDIR = USER_DIR / ".config" / "sshtmux"
T_Host_Style = Literal["panels", "card", "simple", "table", "table2", "json"]
FAST_CONNECTIONS_GROUP_NAME = "fast-connections"
FAST_SESSIONS_NAME = "fast-session"
SFTP_CLI = "sftp"
MULTICOMMNAD_CLI = "multi_command"


class Base(BaseSettings):
    class Config:
        extra = "allow"
        env_file = ".env"


class InternalConfig(BaseModel):
    BASE_DIR: str = str(SSHTMUX_BASEDIR)
    BASE_SERVICE: str = "sshtmux"
    TOML_CONFIG_FILE: str = str(SSHTMUX_BASEDIR / "config.toml")


class SSHTMUX(Base):
    SSHTMUX_IDENTITY_KEY: str | None = None
    SSHTMUX_IDENTITY_KEY_FILE: str | None = str(SSHTMUX_BASEDIR / "identity.key")
    SSHTMUX_IDENTITY_PASSWORDS_FILE: str | None = str(SSHTMUX_BASEDIR / "identity.json")
    SSHTMUX_SNIPPETS_PATH: str | None = str(SSHTMUX_BASEDIR / "snippets")
    SSHTMUX_HOST_STYLE: T_Host_Style = "panels"


class TMUX(Base):
    TMUX_CONFIG_FILE: str = str(SSHTMUX_BASEDIR / "tmux.config")
    TMUX_SOCKET_NAME: str | None = f"sshtmux_{USER_DIR.name}"
    TMUX_SOCKET_PATH: str | None = None
    TMUX_TIMEOUT_COMMANDS: int = 10


class SSH(Base):
    SSH_CONFIG_FILE: str = str(USER_DIR / ".ssh" / "config")
    SSH_COMMAND: str = (
        "ssh -o ConnectTimeout=10 -o StrictHostKeyChecking=no ${hostname}"
    )
    SFTP_COMMAND: str = (
        "sftp -o ConnectTimeout=10 -o StrictHostKeyChecking=no ${hostname}"
    )
    SSH_VALIDATE_SSHCONFIG: str | bool = True
    SSH_CUSTOM_COMMAND: str | bool = False


class ConfigModel(BaseModel):
    internal_config: InternalConfig = InternalConfig()
    sshtmux: SSHTMUX = SSHTMUX()
    ssh: SSH = SSH()
    tmux: TMUX = TMUX()


def update_settings(new_settings):
    global settings
    settings = new_settings


try:
    settings = ConfigModel()
except Exception as e:
    print(f"Init Settings Failed: {e}")
    exit(1)



import re
from typing import Dict, List, Tuple

from pydantic import ValidationError

from ..core.config import T_Host_Style, settings
from ..services.identities import PasswordManager
from .ssh_config import SSH_Config, SSH_Host
from .ssh_parameters import SSHParams


def complete_ssh_host_names(ctx, param, incomplete) -> List[str]:
    ssh_config = SSH_Config().read().parse()
    all_hosts = ssh_config.get_all_host_names()
    return [k for k in all_hosts if k.startswith(incomplete)]


def complete_ssh_group_names(ctx, param, incomplete) -> List[str]:
    ssh_config = SSH_Config().read().parse()
    all_groups = ssh_config.get_all_group_names()
    return [k for k in all_groups if k.startswith(incomplete)]


def complete_params(ctx, param, incomplete) -> List[str]:
    return [k for k in SSHParams.__annotations__.keys() if k.startswith(incomplete)]


def complete_styles(ctx, param, incomplete) -> List[str]:
    return [k for k in T_Host_Style.__args__ if k.startswith(incomplete)]


def complete_identities(ctx, param, incomplete) -> List[str]:
    password_manager = PasswordManager()
    identities = password_manager.get_identities()
    return [k for k in identities if k.startswith(incomplete)]


# We use this functions to give a tuple of group/host names as input, where some names
# can be regexps (with "r:"" prefix), and we evaluate regex based on "all_names" list
# which we use to create a set of expanded host names; direct ones, and expanded ones from
# regex processing. Then return final list...
def expand_names(names: Tuple, all_names: list) -> List[str]:
    selected = set()

    for name in names:
        if name.startswith("r:"):
            name_re = name.split(":")[1]
            # print(f"Got regex type name def - '{name_re}'")
            for i_name in all_names:
                match = re.search(name_re, i_name)
                if match:
                    selected.add(i_name)
        else:
            selected.add(name)
    return list(selected)


# Function to trace connected hosts via "proxyjump" SSH parameter
# For target host, return a list of hosts in order of connections to reach the target host
# In case host is directly reachable (no "proxyjump" param), it will return list of single (target) host
def trace_jumphosts(name: str, config: SSH_Config, ctx, style: str) -> List[SSH_Host]:
    # Keep list of linked hosts via proxyjump option
    traced_hosts: List[SSH_Host] = []

    # we are first checking host that is asked, and then check if that host has a "proxy" defined
    # if proxy is defined, we add found host to traced list, and search attached proxy, it this way we
    # can trace full connection path for later graph processing
    while True:
        # Search for host in current configuration
        if not config.check_host_by_name(name):
            print(
                f"Cannot get info for used host '{name}' as it is not defined in configuration!"
            )
            ctx.exit(1)

        found_host = config.get_host_by_name(name)[0]

        # Set SSH host print style from user input
        found_host.print_style = style

        # Add current host, and its table to traced lists
        traced_hosts.append(found_host)

        # Find proxy info if it exists, if not, break the loop
        proxyjump = None
        host_params = {k.lower(): v for k, v in found_host.params.items()}
        if "proxyjump" in host_params:
            proxyjump = host_params["proxyjump"]
        else:
            for _, params in found_host.inherited_params:
                for key, value in params.items():
                    if key.lower() == "proxyjump":
                        proxyjump = value

        # If there is connected host, we switch "name" and continue the loop
        # Otherwise we exit the loop
        if proxyjump:
            name = proxyjump
        else:
            break

    return traced_hosts


def validate_ssh_params(parameters_input: Tuple[Tuple[str, str]]) -> Dict:
    parameters = {item[0]: item[1] for item in parameters_input}
    if not settings.ssh.SSH_VALIDATE_SSHCONFIG:
        return parameters
    try:
        ssh_params = SSHParams(**parameters)
    except ValidationError as e:
        print("Cannot validate parameters!")
        for error in e.errors():
            loc = error.get("loc")
            if len(loc) > 0:
                print(f"{loc[0]} - {error.get('msg')} ")
            else:
                print(error.get("msg"))
        exit(1)

    valid_ssh_params = ssh_params.model_dump(exclude_none=True)
    return valid_ssh_params


def validate_unique_param(
    host_params: Dict[str, str], new_params: Dict[str, str]
) -> None:
    """
    Remove duplicate Params on SSH config. If there is a keyword in lower case and same with case sensitive.
    Lower will be removed
    """
    host_params_to_delete = []
    for new_param in new_params:
        for host_param in host_params:
            if new_param.lower() == host_param.lower():
                host_params_to_delete.append(host_param)
    for param in host_params_to_delete:
        del host_params[param]
    {**host_params, **new_params}



from textwrap import dedent
from typing import List, Union

from rich import box
from rich.console import Group
from rich.panel import Panel
from rich.table import Table

from ..ssh_host import SSH_Host


# ------------------------------------------------------------------------------
# Render host data in panels with separate sections
# ------------------------------------------------------------------------------
def render(host: SSH_Host):
    out_type = host.type if host.type == "normal" else f"[cyan]{host.type}[/]"
    out_info = "\n".join(host.info) if host.info else "- No info defined - "

    grp_inputs: List[Union[Table, Panel]] = []

    # // Add Host data panel to the group
    # // -----------------------------------------------------------------------
    host_panel_data = f"""\
    [bright_white]Name [/]:  {host.name}
    [bright_white]Group[/]:  {host.group}
    [bright_white]Type [/]:  {out_type}\
    """
    grp_inputs.append(
        Panel(
            dedent(host_panel_data),
            border_style="grey35",
            title="Host",
            title_align="left",
        )
    )

    # // Add Host info panel to the group (if host info exist)
    # // -----------------------------------------------------------------------
    if host.info:
        grp_inputs.append(
            Panel(
                out_info,
                border_style="grey35",
                style="gray50",
                title="Info",
                title_align="left",
            )
        )

    # // Prepare table with params and append it to the group
    # // -----------------------------------------------------------------------
    param_table = Table(
        box=box.ROUNDED, style="grey35", show_header=True, show_edge=True, expand=True
    )
    param_table.add_column("Param")
    param_table.add_column("Value")
    param_table.add_column("Inherited-from")

    # Add rows for SSH Config parameter table
    for key, value in host.params.items():
        output_value = value if not isinstance(value, list) else "\n".join(value)
        param_table.add_row(key, output_value)

    # Add rows for inherited SSH Config parameters
    for pattern, pattern_params in host.inherited_params:
        for param, value in pattern_params.items():
            if param not in host.params:
                output_value = (
                    value if not isinstance(value, list) else "\n".join(value)
                )
                param_table.add_row(param, output_value, pattern, style="yellow")

    grp_inputs.append(param_table)

    # // Render output
    # // -----------------------------------------------------------------------
    return Panel.fit(Group(*grp_inputs), box=box.SIMPLE, padding=(0, 0))



from rich import box
from rich.console import Group
from rich.panel import Panel
from rich.table import Table

from ..ssh_host import SSH_Host


# ------------------------------------------------------------------------------
# Render host data in panel without borders and minimal decorations
# ------------------------------------------------------------------------------
def render(host: SSH_Host):
    out_type = host.type if host.type == "normal" else f"[cyan]{host.type}[/]"

    # // Add Host data information
    # // -----------------------------------------------------------------------
    panel_data = [
        f"[bright_white]Name [/]:  {host.name}",
        f"[bright_white]Group[/]:  {host.group}",
        f"[bright_white]Type [/]:  {out_type}",
    ]

    # // Add Host info data (if host info exist)
    # // -----------------------------------------------------------------------
    if host.info:
        out_info = "\n".join(host.info)
        panel_data = panel_data + [
            "",
            f"[gray50]{out_info}[/]",
        ]
    host_panel = Panel(
        "\n".join(panel_data), box=box.SIMPLE, border_style="grey35", padding=(0, 0)
    )

    # // Prepare table with params
    # // -----------------------------------------------------------------------
    param_table = Table(
        box=box.SIMPLE, style="grey35", show_header=True, show_edge=False, pad_edge=True
    )
    param_table.add_column("Param")
    param_table.add_column("Value")
    param_table.add_column("Inherited-from")

    # Add rows for SSH Config parameters
    for key, value in host.params.items():
        output_value = value if not isinstance(value, list) else "\n".join(value)
        param_table.add_row(key, output_value)

    # Add rows for inherited SSH Config parameters
    for pattern, pattern_params in host.inherited_params:
        for param, value in pattern_params.items():
            if param not in host.params:
                output_value = (
                    value if not isinstance(value, list) else "\n".join(value)
                )
                param_table.add_row(param, output_value, pattern, style="yellow")

    param_table.add_row("")

    # // Render output
    # // -----------------------------------------------------------------------
    return Group(host_panel, param_table)



import click

from sshtmux.sshm import SSH_Config, complete_ssh_group_names

# TODO: Check click.edit for multiline edit option (info, or even params?)

# ------------------------------------------------------------------------------
# COMMAND: group set
# ------------------------------------------------------------------------------
SHORT_HELP = "Change group parameters"
LONG_HELP = """
Change/modify group parameters

Command allows to modify current information on the group, like info lines
and description. To rename group, use command "sshm group rename" command.
To move hosts between group, use "sshm host set" command.
"""

# Parameters help:
DESC_HELP = "Short description of group"
INFO_HELP = "Info line, can be set multiple times"
# ------------------------------------------------------------------------------


@click.command(name="set", short_help=SHORT_HELP, help=LONG_HELP)
@click.option("-d", "--desc", help=DESC_HELP)
@click.option("-i", "--info", multiple=True, help=INFO_HELP)
@click.argument("name", shell_complete=complete_ssh_group_names)
@click.pass_context
def cmd(ctx, name, desc, info):
    config: SSH_Config = ctx.obj

    # Nothing was provided
    if not desc and not info:
        print("Calling set on group without specifying any option is not valid!")
        print("Run 'sshm group set -h' for help.")
        ctx.exit(1)

    if not config.check_group_by_name(name):
        print(f"Cannot modify group '{name}', it is not defined in configuration!")
        ctx.exit(1)

    found_group = config.get_group_by_name(name)

    # If new description is set
    if desc:
        found_group.desc = desc.strip()

    # Assuming when "clearing" config only single info is given... we check first info
    # and check if it's empty/falsy. If it is non-empty, we check all of values and if
    # they are non empty we append to existing lines.... if first item is empty we clear the info data
    if info and info[0]:
        for line in info:
            if len(line.strip()) > 0:
                found_group.info.append(line)
    else:
        found_group.info = []

    config.generate_ssh_config().write_out()

    if not config.stdout:
        print(f"Modified group: {name}")



CONNECTIONS_ERRORS = [
    "failed",
    "denied",
    "error",
    "closed",
    "invalid",
    "disabled",
    "Bad",
    "Bad configuration",
    "refused",
    "failure",
    "Could not",
    "Connection reset by peer",
    "Received disconnect",
    "No supported authentication methods available",
    "Unable to negotiate with",
    "Connection timed out",
    "Broken pipe",
    "Network is unreachable",
    "Host is down",
    "Connection reset by [host]",
    "Resource temporarily unavailable",
    "Protocol major versions differ",
    "Protocol mismatch",
    "Connection aborted",
    "SSH protocol version mismatch",
    "Host identification has changed",
    "Operation timed out",
    "Host is unreachable",
    "No route to host",
    "Connection aborted by peer",
    "SSH packet corrupt",
    "Connection lost",
    "Broken authentication sequence",
    "No space left on device",
    "Host not found",
    "Address already in use",
    "Connection timed out waiting for session",
    "Host public key mismatch",
    "Cannot bind to port",
    "Unsupported protocol version",
    "Unexpected EOF in input",
    "SSH connection dropped",
    "Server host key mismatch",
    "Authentication rejected by peer",
    "Host key does not match",
    "Max sessions reached",
    "Too many concurrent sessions",
    "Unexpected disconnection",
    "Host is blacklisted",
    "User not allowed by policy",
    "Received SSH_MSG_DISCONNECT",
    "SSH key expired",
    "Connection attempt rejected",
    "SSH agent not available",
    "RSA host key for has changed",
    "Fingerprint mismatch",
    "Unable to connect",
    "Host temporarily unavailable",
    "No response from host",
    "Read from socket failed: Connection timed out",
    "Timeout waiting for input",
    "Timeout before authentication",
    "Session timeout",
    "Client timeout",
    "Timeout while connecting",
    "Timeout expired",
    "Handshake timeout",
    "Operation timeout",
    "Login timeout",
    "Network timeout",
    "SSH connection timeout",
    "Request timeout",
    "Timed out waiting for response",
    "Connection timeout during handshake",
    "Timeout error during key exchange",
    "Socket timeout",
]



import os
import time
from abc import ABC, abstractmethod
from enum import Enum
from typing import List, Union

import libtmux
from libtmux import Window
from rich import print
from rich.prompt import Prompt

from sshtmux.core.config import (
    FAST_CONNECTIONS_GROUP_NAME,
    FAST_SESSIONS_NAME,
    MULTICOMMNAD_CLI,
    SFTP_CLI,
    settings,
)
from sshtmux.exceptions import IdentityException, SSHException, TMUXException
from sshtmux.services.connections_erros import CONNECTIONS_ERRORS
from sshtmux.services.identities import PasswordManager, prompt_identity
from sshtmux.services.snippets import prompt_snippet
from sshtmux.sshm import SSH_Config, SSH_Host


class ConnectionAbstract(ABC):
    def __init__(self) -> None:
        super().__init__()
        self.password_manager = PasswordManager()
        self.connection_cmd = settings.ssh.SSH_COMMAND

    @abstractmethod
    def start(
        self,
        window: Window,
        host: SSH_Host,
        identity: Union[str, None],
    ):
        """
        Start SSH connection
        """

    def _check_connections_errors(
        self,
        window: Window,
        pane_output: Union[str, List[str]],
        host: SSH_Host,
    ):
        if isinstance(pane_output, str):
            pane_output = [pane_output]

        is_failed = False
        for text in pane_output:
            if text.startswith("ssh:") or text.startswith(
                "ssh_exchange_identification:"
            ):
                is_failed = True
            else:
                for error in CONNECTIONS_ERRORS:
                    if error.lower() in text.lower():
                        is_failed = True
                        break

            if is_failed:
                window.kill_window()
                error_msg = (
                    pane_output[-2] if len(pane_output) >= 2 else str(pane_output)
                )
                raise SSHException(f"{host}\n\nConnection Error!\n\n{error_msg}")

    def _check_timeout_reached(
        self,
        timeout_start,
        pane_output: List[str],
        host: SSH_Host,
        window: Window,
    ):
        if time.time() - timeout_start > settings.tmux.TMUX_TIMEOUT_COMMANDS:
            timeout_mgs = f"{host}\n\n Timeout reached!"
            window.kill_window()
            if pane_output and len(pane_output) >= 2:
                error_msg = (
                    pane_output[-2] if len(pane_output) >= 2 else str(pane_output)
                )
                raise SSHException(f"{timeout_mgs} \n\n {error_msg}")
            raise SSHException(f"{timeout_mgs} \n\n {host} \n\n {pane_output}")


class NormalConnection(ConnectionAbstract):
    def start(
        self,
        window: Window,
        host: SSH_Host,
        identity: Union[str, None],
    ):
        pane_output = None
        cmd = self.connection_cmd.replace("${hostname}", host.name)
        window.attached_pane.send_keys(cmd)
        timeout_start = time.time()
        password_prompt_found = False
        while not password_prompt_found:
            self._check_timeout_reached(timeout_start, pane_output, host, window)
            pane_output = window.attached_pane.capture_pane()
            self._check_connections_errors(window, pane_output, host)

            for text in pane_output:
                if "password:" in text:
                    password_prompt_found = True
                    break
            time.sleep(0.1)


class IdentityConnection(ConnectionAbstract):
    def start(
        self,
        window: Window,
        host: SSH_Host,
        identity: Union[str, None],
    ):
        pane_output = None
        cmd = self.connection_cmd.replace("${hostname}", host.name)
        try:
            password = self.password_manager.get_password(identity)
        except IdentityException as e:
            window.kill_window()
            raise e
        window.attached_pane.send_keys(cmd)

        timeout_start = time.time()
        password_prompt_found = False
        while not password_prompt_found:
            self._check_timeout_reached(timeout_start, pane_output, host, window)

            pane_output = window.attached_pane.capture_pane()
            self._check_connections_errors(window, pane_output, host)

            for text in pane_output:
                if "password:" in text:
                    password_prompt_found = True
                    break
            time.sleep(0.1)

        if password_prompt_found:
            window.attached_pane.send_keys(password)
            self._validate_ssh_session_identity(window, host, pane_output)

    def _validate_ssh_session_identity(
        self,
        window: Window,
        host: SSH_Host,
        last_pane_output: List[str],
    ):
        pane_output = None
        timeout_start = time.time()

        while True:
            self._check_timeout_reached(timeout_start, pane_output, host, window)

            pane_output = window.attached_pane.capture_pane()
            if last_pane_output == pane_output:
                time.sleep(0.1)
                continue
            self._check_connections_errors(window, pane_output, host)
            break


class CustomConnection(ConnectionAbstract):
    def start(
        self,
        window: Window,
        host: SSH_Host,
        identity: Union[str, None],
    ):
        if not settings.ssh.SSH_CUSTOM_COMMAND:
            raise TMUXException("SSH_CUSTOM_COMMAND not set")

        cmd = settings.ssh.SSH_CUSTOM_COMMAND.replace("${hostname}", host.name)
        if identity:
            try:
                password = self.password_manager.get_password(identity)
            except IdentityException as e:
                window.kill_window()
                raise e
            cmd = cmd.replace("${password}", password)
        else:
            cmd = cmd.replace("${password}", "")
        window.attached_pane.send_keys(cmd)


class SFTPNormalConnection(NormalConnection):
    def __init__(self):
        super().__init__()
        self.connection_cmd = settings.ssh.SFTP_COMMAND


class SFTPIdentityConnection(IdentityConnection):
    def __init__(self):
        super().__init__()
        self.connection_cmd = settings.ssh.SFTP_COMMAND


class ConnectionProtocol(Enum):
    ssh = "SSH"
    sftp = "SFTP"


class ConnectionType(Enum):
    normal = NormalConnection
    identity = IdentityConnection
    custom = CustomConnection
    sftp_normal = SFTPNormalConnection
    sftp_identity = SFTPIdentityConnection


class Tmux:
    def __init__(self) -> None:
        self.server = libtmux.Server(
            socket_name=settings.tmux.TMUX_SOCKET_NAME,
            socket_path=settings.tmux.TMUX_SOCKET_PATH,
            config_file=settings.tmux.TMUX_CONFIG_FILE,
        )

    def create_window(
        self,
        type_connection: ConnectionType,
        host: SSH_Host,
        attach=False,
        identity: Union[str, None] = None,
        overwritten_group: str = None,
    ):
        window = None
        window_name = host.name
        session_name = host.group
        connection: ConnectionAbstract = type_connection.value()
        if overwritten_group:
            session_name = overwritten_group
        session = self.server.find_where({"session_name": session_name})

        if window_name.startswith(session_name):
            window_name = window_name.replace(f"{session_name}-", "")

        if not session:
            self.server.cmd(
                "new-session", "-d", "-P", "-F#{session_id}", "-s", session_name
            )
            session = self.server.find_where({"session_name": session_name})
            if not session:
                raise TMUXException(f"{host}\n\n Something went wrong to find session!")

            if len(session.windows) == 1:
                window = session.windows[0].rename_window(window_name)
            else:
                raise TMUXException(
                    f"{host}\n\n Something went wrong to find inicial window!"
                )
        else:
            window = session.new_window(window_name=window_name, attach=attach)

        if window:
            try:
                connection.start(window, host, identity)
            except KeyboardInterrupt as e:
                raise TMUXException(str(e))
        else:
            raise TMUXException(f"{host}\n\n Something went wrong to find window!")

        if attach:
            session.attach()
        return True

    def attach(self):
        sessions = self.server.list_sessions()

        if sessions:
            current_session = sessions[0]

            if current_session:
                current_session.cmd("choose-session")
                current_session.attach()
                return True
        return False

    def execute_cmd_tmux(self, cmd, session_name, window_index, panel_index):
        session = self.server.find_where({"session_name": session_name})
        if not session:
            print("No Tmux session available")
            return
        window = [w for w in session.windows if str(w.index) == str(window_index)][0]
        panel = [p for p in window.panes if str(p.index) == str(panel_index)][0]
        panel.send_keys(cmd)

    def execute_cmd_shell(self, cmd):
        if not cmd:
            return
        cmd = cmd.split()
        os.execvp(cmd[0], cmd)

    def execute_snippet(self, session_name, window_index, panel_index):
        snippet = prompt_snippet()
        if not snippet:
            print("No Snippets found")
            return

        is_tmux = all([session_name, window_index, panel_index])
        if is_tmux:
            self.execute_cmd_tmux(snippet, session_name, window_index, panel_index)
        else:
            self.execute_cmd_shell(snippet)

    def execute_identity(self, session_name, window_index, panel_index):
        identity = prompt_identity()
        if not identity:
            print("No Identity found")
            return
        self.execute_cmd_tmux(identity, session_name, window_index, panel_index)

    def execute_multi_command(self, session):
        cmd = Prompt.ask("Multi Session Command")
        if not cmd:
            return
        for window in session.windows:
            for panel in window.panes:
                panel.send_keys(cmd)

    def execute_host_cmd(self, session_name, window_index, panel_index, cmd_ref):
        cmd = None
        session = self.server.find_where({"session_name": session_name})
        if not session:
            print("No Tmux session available")
            return

        window = [w for w in session.windows if str(w.index) == str(window_index)][0]
        if session_name in [
            FAST_CONNECTIONS_GROUP_NAME,
            FAST_SESSIONS_NAME,
            SSH_Config.DEFAULT_GROUP_NAME,
        ]:
            hostname = window.name
        else:
            if not window.name.startswith(f"{session_name}-"):
                hostname = f"{session_name}-{window.name}"
            else:
                hostname = window.name

        if cmd_ref == SFTP_CLI:
            cmd = (
                settings.ssh.SFTP_COMMAND.replace("${hostname}", hostname)
                .replace("&& exit", "")
                .replace("; exit", "")
            )
            self.execute_cmd_shell(cmd)

        if cmd_ref == MULTICOMMNAD_CLI:
            self.execute_multi_command(session)



import curses
import os
from time import sleep

from rich.console import Console
from rich.prompt import Prompt

from sshtmux.core.config import settings

console = Console()


def choose_cmd(path):
    lines = None

    try:
        with open(path, "r") as arquivo:
            lines = arquivo.readlines()
    except FileNotFoundError:
        print("File not found")
        exit(1)
    except Exception as e:
        print(f"Error to open file {path}: {e}")
        exit(1)

    if not lines:
        print(f"Empty file: {path}")
        return None

    def browser(stdscr):
        index = 0
        scroll = 0
        height, width = stdscr.getmaxyx()

        while True:
            stdscr.clear()
            stdscr.addstr("Choose a command\n\n")

            for i in range(scroll, min(scroll + height - 3, len(lines))):
                line_showed = lines[i].strip()
                if len(line_showed) > width - 4:
                    line_showed = line_showed[: width - 4]

                if i == index:
                    stdscr.addstr(f">> {i + 1}: {line_showed}\n", curses.A_REVERSE)
                else:
                    stdscr.addstr(f"   {i + 1}: {line_showed}\n")

            key = stdscr.getch()

            if key == ord("j") or key == curses.KEY_DOWN:
                if index < len(lines) - 1:
                    index += 1
                if index >= scroll + height - 3:
                    scroll += 1
            elif key == ord("k") or key == curses.KEY_UP:
                if index > 0:
                    index -= 1
                if index < scroll:
                    scroll -= 1

            elif key == ord("\n"):
                return lines[index].strip()

            elif key == ord("q"):
                break

    return curses.wrapper(browser)


def get_snippets_files():
    snippets = ["Cancel"]
    snippets_path = settings.sshtmux.SSHTMUX_SNIPPETS_PATH
    if os.path.exists(snippets_path):
        for _, _, filenames in os.walk(snippets_path):
            for filename in filenames:
                snippets.append(filename)
    snippets_idx = [str(index) for index, _ in enumerate(snippets)]
    return snippets, snippets_idx


def _show_menu(options):
    console.print("Snippets:", style="bold underline")
    for idx, option in enumerate(options):
        console.print(f"{idx}. {option}")


def prompt_snippet():
    cmd = None
    choices, choices_idx = get_snippets_files()
    _show_menu(choices)
    choice = Prompt.ask(
        "Choose snippet:",
        choices=choices_idx,
        default=choices_idx[0],
        show_choices=False,
    )
    snippet_file = choices[int(choice)]
    if snippet_file == choices[0]:
        return

    try:
        cmd = choose_cmd(
            f"{settings.sshtmux.SSHTMUX_SNIPPETS_PATH}{os.sep}{snippet_file}"
        )
    except Exception as e:
        print(str(e))
        sleep(2)
    return cmd



import click

from sshtmux.sshm import SSH_Config, complete_ssh_group_names

# ------------------------------------------------------------------------------
# COMMAND: group rename
# ------------------------------------------------------------------------------
SHORT_HELP = "Rename existing group"
LONG_HELP = "Rename existing group in configuration"

# ------------------------------------------------------------------------------


@click.command(name="rename", short_help=SHORT_HELP, help=LONG_HELP)
@click.argument("name", shell_complete=complete_ssh_group_names)
@click.argument("new_name")
@click.pass_context
def cmd(ctx, name, new_name):
    config: SSH_Config = ctx.obj

    if not config.check_group_by_name(name):
        print(f"Cannot rename group '{name}', as it is not defined in configuration!")
        ctx.exit(1)
    if config.check_group_by_name(new_name):
        print(
            f"Cannot rename group '{name}' to '{new_name}' as new name is already used!"
        )
        ctx.exit(1)

    config.get_group_by_name(name).name = new_name
    config.generate_ssh_config().write_out()

    if not config.stdout:
        print(f"Renamed group: {name} -> {new_name}")



from rich import box
from rich.panel import Panel
from rich.table import Table

from ..ssh_host import SSH_Host


# ------------------------------------------------------------------------------
# Render host data in panel with single-card look
# ------------------------------------------------------------------------------
def render(host: SSH_Host):
    out_type = host.type if host.type == "normal" else f"[cyan]{host.type}[/]"

    # // Add Host data information
    # // -----------------------------------------------------------------------
    layout_table = Table(
        "", box=box.ROUNDED, style="grey35", show_edge=True, show_header=False
    )
    layout_table.add_row(f"[bright_white]Name [/]:  {host.name}")
    layout_table.add_row(f"[bright_white]Group[/]:  {host.group}")
    layout_table.add_row(f"[bright_white]Type [/]:  {out_type}")

    # // Add Host info panel (if host info exist)
    # // -----------------------------------------------------------------------
    if host.info:
        layout_table.add_row("")
        out_info = "\n".join(host.info)
        layout_table.add_row(f"[gray50]{out_info}[/]")

    layout_table.add_row("")

    # // Prepare table with params and append it to the layout table
    # // -----------------------------------------------------------------------
    param_table = Table(
        box=box.SIMPLE,
        style="grey35",
        show_header=True,
        show_edge=False,
        pad_edge=False,
    )
    param_table.add_column("Param")
    param_table.add_column("Value")
    param_table.add_column("Inherited-from")

    # Add rows for SSH Config parameters
    for key, value in host.params.items():
        output_value = value if not isinstance(value, list) else "\n".join(value)
        param_table.add_row(key, output_value)

    # Add rows for inherited SSH Config parameters
    for pattern, pattern_params in host.inherited_params:
        for param, value in pattern_params.items():
            if param not in host.params:
                output_value = (
                    value if not isinstance(value, list) else "\n".join(value)
                )
                param_table.add_row(param, output_value, pattern, style="yellow")

    layout_table.add_row(param_table)

    return Panel(layout_table, box=box.SIMPLE, padding=(0, 0))



from typing import ClassVar

from rich import box
from rich.panel import Panel
from rich.rule import Rule
from rich.table import Table
from textual.app import App, ComposeResult
from textual.binding import Binding, BindingType
from textual.containers import Container, VerticalScroll
from textual.widgets import (
    ContentSwitcher,
    Footer,
    Header,
    Input,
    Label,
    OptionList,
    Static,
    Tree,
)
from textual.widgets.option_list import Separator

from sshtmux.core.config import (
    FAST_CONNECTIONS_GROUP_NAME,
    FAST_SESSIONS_NAME,
    settings,
)
from sshtmux.exceptions import IdentityException, SSHException, TMUXException
from sshtmux.services.identities import PasswordManager
from sshtmux.services.tmux import ConnectionProtocol, ConnectionType, Tmux
from sshtmux.sshm import SSH_Config, SSH_Group, SSH_Host


class SSHGroupDataInfo(Static):
    """Widget for SSH Group data"""

    DEFAULT_CSS = """
    .grp_labels {
        width: 100%;
        color: $accent;
        padding: 1 1 0 1;
    }
    """

    def compose(self) -> ComposeResult:
        yield Label("Description", classes="grp_labels")
        yield Static(Panel("...empty...", style="grey42"), id="grp_description")
        yield Label("Extra information", classes="grp_labels")
        yield Static(Panel("...empty...", style="grey42"), id="grp_information")

    def update(self, group: SSH_Group) -> None:
        desc: Static = self.query_one("#grp_description")  # type: ignore
        info: Static = self.query_one("#grp_information")  # type: ignore
        desc.update(Panel(group.desc, border_style="grey42"))
        info.update(Panel("\n".join(group.info), border_style="grey42"))


class SSHHostDataInfo(Static):
    """Widget for SSH Host data"""

    DEFAULT_CSS = """
    .hst_labels {
        width: 100%;
        color: $accent;
        padding: 1 1 0 1;
    }
    """

    def compose(self) -> ComposeResult:
        yield Label("Details", classes="hst_labels")
        yield Static(Panel("...empty...", style="grey42"), id="hst_details")
        yield Label("Extra information", classes="hst_labels")
        yield Static(Panel("...empty...", style="grey42"), id="hst_information")
        yield Label("SSH Parameters", classes="hst_labels")
        yield Static(Panel("...empty...", style="grey42"), id="hst_parameters")

    def update(self, host: SSH_Host) -> None:
        det: Static = self.query_one("#hst_details")  # type: ignore
        info: Static = self.query_one("#hst_information")  # type: ignore
        params: Static = self.query_one("#hst_parameters")  # type: ignore
        det.update(
            Panel(
                f"[bold]Group[/b]: {host.group}\n[bold]Type[/b]:  {host.type}",
                border_style="grey42",
            )
        )
        info.update(
            Panel("\n".join(host.info), border_style="grey42")
            if host.info
            else Panel("...empty...", style="grey42")
        )

        param_table = Table(
            box=box.ROUNDED,
            style="grey42",
            show_header=True,
            show_edge=True,
            expand=True,
        )
        param_table.add_column("Param")
        param_table.add_column("Value")
        param_table.add_column("Inherited-from")

        # Add rows for SSH Config parameter table
        for key, value in host.params.items():
            output_value = value if not isinstance(value, list) else "\n".join(value)
            param_table.add_row(key, output_value)

        # Add rows for inherited SSH Config parameters
        for pattern, pattern_params in host.inherited_params:
            for param, value in pattern_params.items():
                if param not in host.params:
                    output_value = (
                        value if not isinstance(value, list) else "\n".join(value)
                    )
                    param_table.add_row(param, output_value, pattern, style="yellow")

        params.update(param_table)  # type: ignore


class SSHDataView(Static):
    """SSH Item data view panel"""

    def compose(self) -> ComposeResult:
        yield Label(id="data_view_header")
        yield Static(Rule(style="$text"))

        with ContentSwitcher(initial="no-view"):
            yield Static(id="no-view")
            yield SSHGroupDataInfo(id="group-view")
            with VerticalScroll(id="host-view"):
                yield SSHHostDataInfo()

    def update(self, sshitem="") -> None:
        label: Label = self.query_one("#data_view_header")  # type: ignore
        grp = self.query_one(SSHGroupDataInfo)
        hst = self.query_one(SSHHostDataInfo)

        if isinstance(sshitem, SSH_Group):
            group: SSH_Group = sshitem
            label.update(f"Group: {group.name}")
            self.query_one(ContentSwitcher).current = "group-view"
            grp.update(group)

        elif isinstance(sshitem, SSH_Host):
            host: SSH_Host = sshitem
            label.update(f"Host: {host.name}")
            self.query_one(ContentSwitcher).current = "host-view"
            hst.update(host)

        else:
            label.update("Select node from the list")
            self.query_one(ContentSwitcher).current = "no-view"


class CustomOptionList(OptionList):
    BINDINGS: ClassVar[list[BindingType]] = [
        Binding("down", "cursor_down", "Down", show=False),
        Binding("j", "cursor_down", "Down", show=False),
        Binding("end", "last", "Last", show=False),
        Binding("enter", "select", "Select", show=False),
        Binding("home", "first", "First", show=False),
        Binding("pagedown", "page_down", "Page down", show=False),
        Binding("pageup", "page_up", "Page up", show=False),
        Binding("up", "cursor_up", "Up", show=False),
        Binding("k", "cursor_up", "Up", show=False),
    ]


class SSHTui(App):
    TITLE = "SSHTMUX"

    BINDINGS = [
        Binding("q", "quit", "Quit"),
        Binding("t", "attach_tmux", "TMUX"),
        Binding("d", "detached_ssh", "Detached SSH"),
        Binding("c", "connect_ssh", "Conect SSH"),
        Binding("s", "connect_sftp", "SFTP to host"),
        Binding("f", "connect_fast_connections", "Fast Connection"),
        Binding("F", "connect_fast_session", "Fast Session"),
        Binding("m", "toggle_dark", "Switch background mode", False),
        Binding("j", "cursor_down", "Cursor Down", False),
        Binding("k", "cursor_up", "Cursor Up", False),
        Binding("l", "cursor_expand", "Expand Node Tree", False),
        Binding("h", "cursor_collapse", "Collapse Node Tree", False),
        Binding("?", "search_groups", "Search Groups"),
        Binding("/", "search_hosts", "Search Hosts"),
        Binding("escape", "clean_filters", "Clear all filters", False),
    ]

    CSS = """
    Screen {
        background: $surface-darken-1;
    }

    SSHDataView {
        width: 100%;
        height: 100%;
        padding: 1 2;
        background: $panel;
        content-align: center middle;
    }

    Tree {
        scrollbar-gutter: stable;
        overflow: auto;
        width: 36;
        height: 100%;
        dock: left;
        background: $surface;
    }
    """

    def __init__(self, sshmconf=None):
        self.tmux = Tmux()
        self.password_manager = PasswordManager()
        self.identities = self.password_manager.get_identities()
        self.attach_connection = False
        self.connections_tree = None
        self.overwritten_group = None
        if isinstance(sshmconf, SSH_Config):
            self.sshmconf = sshmconf
        else:
            self.sshmconf = SSH_Config().read().parse()

        super().__init__()

    def compose(self) -> ComposeResult:
        yield Header(show_clock=True)
        with Container():
            self.connections_tree = Tree(
                f"SSH Configuration ({len(self.sshmconf.groups)} groups)",
                id="sshtree",
                data=None,
            )
            self._generate_tree()

            yield self.connections_tree
            yield SSHDataView()

        self.input_groups_search = Input(
            placeholder="Search Groups...", id="search_groups_input"
        )
        self.input_groups_search.display = False
        self.input_hosts_search = Input(
            placeholder="Search Hosts...", id="search_hosts_input"
        )
        self.input_hosts_search.display = False
        self.input_fast_connections = Input(
            placeholder="user@hostname", id="fast_connections_input"
        )
        self.input_fast_connections.display = False
        self.type_password = "Type Password"
        self.select_identity = CustomOptionList(
            self.type_password, Separator(), *self.identities, id="select_identity"
        )
        self.select_identity.display = False
        yield self.input_groups_search
        yield self.input_hosts_search
        yield self.select_identity
        yield self.input_fast_connections
        yield Footer()

    async def on_option_list_option_selected(
        self, event: OptionList.OptionSelected
    ) -> None:
        """
        Manager identity
        """
        value = event.option.prompt

        if not self._is_sshhost():
            return

        self.connections_tree.focus()
        self.select_identity.display = False

        if value == self.type_password:
            if self.connection == ConnectionProtocol.ssh:
                self._start_connection(ConnectionType.normal, None)
            elif self.connection == ConnectionProtocol.sftp:
                self._start_connection(ConnectionType.sftp_normal, None)
        elif value:
            if self.connection == ConnectionProtocol.ssh:
                self._start_connection(ConnectionType.identity, value)
            elif self.connection == ConnectionProtocol.sftp:
                self._start_connection(ConnectionType.sftp_identity, value)

    def on_mount(self, _) -> None:
        self.ENABLE_COMMAND_PALETTE = False
        self.connections_tree.focus()

    def on_tree_node_highlighted(self, event):
        self.current_node = event.node.data
        self.query_one(SSHDataView).update(self.current_node)

    def action_toggle_dark(self) -> None:
        self.dark = not self.dark

    def action_quit(self) -> None:
        self.exit(0)

    def action_search_groups(self) -> None:
        self.input_hosts_search.value = ""
        self.input_hosts_search.display = False
        self.input_groups_search.display = True
        self.input_groups_search.focus()

    def action_search_hosts(self) -> None:
        self.input_groups_search.value = ""
        self.input_groups_search.display = False
        self.input_hosts_search.display = True
        self.input_hosts_search.focus()

    def action_connect_fast_connections(self) -> None:
        self.input_fast_connections.display = True
        self.input_fast_connections.focus()

    def action_connect_fast_session(self) -> None:
        self.overwritten_group = FAST_SESSIONS_NAME
        self.action_connect_ssh(attach=True)

    def action_clean_filters(self) -> None:
        self.input_fast_connections.value = ""
        self.input_groups_search.display = False
        self.input_hosts_search.display = False
        self.select_identity.display = False
        self.input_fast_connections.display = False
        self.connections_tree.focus()

    def action_cursor_down(self) -> None:
        if self.connections_tree.cursor_line == -1:
            self.connections_tree.cursor_line = 0
        else:
            self.connections_tree.cursor_line += 1
        self.connections_tree.scroll_to_line(self.connections_tree.cursor_line)

    def action_cursor_up(self) -> None:
        if self.connections_tree.cursor_line == -1:
            self.connections_tree.cursor_line = self.connections_tree.last_line
        else:
            self.connections_tree.cursor_line -= 1
        self.connections_tree.scroll_to_line(self.connections_tree.cursor_line)

    def action_cursor_expand(self) -> None:
        self.connections_tree.cursor_node.expand()

    def action_cursor_collapse(self) -> None:
        self.connections_tree.cursor_node.collapse()

    def action_attach_tmux(self) -> None:
        attached = self._run_external_func_with_args(self.tmux.attach)
        if not attached:
            self.notify("No Tmux session available", severity="warning")

    def action_connect_ssh(self, attach=True):
        if not self._is_sshhost():
            return

        self.connection = ConnectionProtocol.ssh
        self.atatch_connection = attach
        if not self.identities:
            self._start_connection(ConnectionType.normal, None)
            return
        self.select_identity.display = True
        self.select_identity.focus()

    def action_detached_ssh(self) -> None:
        self.connection = ConnectionProtocol.ssh
        self.action_connect_ssh(False)

    def action_connect_sftp(self) -> None:
        self.connection = ConnectionProtocol.sftp
        self.atatch_connection = True
        if not self._is_sshhost():
            return
        if not self.identities:
            self._start_connection(ConnectionType.sftp_normal, None)
            return
        self.select_identity.display = True
        self.select_identity.focus()

    def on_input_submitted(self, event: Input.Submitted) -> None:
        event.input.display = False
        event_id = event.input.id
        if event_id == "search_groups_input" or event_id == "search_hosts_input":
            self.connections_tree.focus()
        elif event_id == "fast_connections_input":
            self._fast_connections_input_submit(event)

    def on_input_changed(self, event: Input.Changed) -> None:
        event_id = event.input.id
        if event_id == "search_groups_input" or event_id == "search_hosts_input":
            self._search_input_changed(event)

    def _search_input_changed(self, event: Input.Changed):
        filter = event.value
        self.connections_tree.clear()
        if event.input.id == "search_groups_input":
            self._generate_tree(filter_groups=filter)
        elif event.input.id == "search_hosts_input":
            self._generate_tree(filter_hosts=filter)
            for node in self.connections_tree.root.children:
                node.expand()
                for child in node.children:
                    child.expand()

        if filter == "":
            for node in self.connections_tree.root.children:
                node.collapse_all()

    def _fast_connections_input_submit(self, event: Input.Submitted):
        self.connections_tree.focus()
        if event.value == "":
            event.input.display = False
            return

        host = event.value
        params = {}
        host_parts = host.split("@")
        if len(host_parts) >= 2:
            params["user"] = host_parts[0]
        self.atatch_connection = True
        self.current_node = SSH_Host(
            name=host,
            params=params,
            group=FAST_CONNECTIONS_GROUP_NAME,
        )
        self.action_connect_ssh(attach=True)

    def _generate_tree(
        self, *, filter_hosts: str | None = None, filter_groups: str | None = None
    ):
        self.connections_tree.root.expand()

        groups = self.sshmconf.groups_sorted
        if filter_hosts:
            groups_filtered = []
            for group in groups:
                hosts = [
                    h
                    for h in group.hosts
                    if filter_hosts in h.name.replace(group.name, "")
                ]
                if len(hosts) > 0:
                    group.hosts = hosts
                    groups_filtered.append(group)
            groups = groups_filtered

        elif filter_groups:
            groups = [g for g in groups if filter_groups in g.name]

        for group in groups:
            g = self.connections_tree.root.add(
                f":file_folder: {group.name}", data=group, expand=False
            )
            for host in group.hosts + group.patterns:
                g.add_leaf(host.name, data=host)

    def _is_sshhost(self):
        if not (
            isinstance(self.current_node, SSH_Host)
            and self.current_node.type == "normal"
        ):
            self.notify("Only normal hosts connections are allowed", severity="warning")
            return False
        return True

    def _start_connection(self, type_connection, identity):
        if not self._is_sshhost():
            return

        if settings.ssh.SSH_CUSTOM_COMMAND:
            type_connection = ConnectionType.custom

        is_conneted = self._run_external_func_with_args(
            self.tmux.create_window,
            type_connection=type_connection,
            host=self.current_node,
            attach=self.atatch_connection,
            identity=identity,
            overwritten_group=self.overwritten_group,
        )
        self.overwritten_group = None

        if is_conneted:
            self.notify(
                f"Connected to:\n\n{self.current_node}",
                severity="information",
            )

    def _run_external_func_with_args(self, func, **kwargs):
        driver = self._driver
        result = None
        if driver is not None:
            driver.stop_application_mode()
            try:
                result = func(**kwargs)
            except TMUXException as e:
                self.notify(str(e), title="Tmux", severity="error")
            except SSHException as e:
                self.notify(str(e), title="SSH", severity="error")
            except IdentityException as e:
                self.notify(str(e), title="Identity", severity="error")
            except Exception as e:
                known_errors = [
                    "no server running on",
                    "could not find object",
                ]
                if any(error in str(e).lower() for error in known_errors):
                    pass
                elif "sessions should be nested with care" in str(e):
                    self.notify(
                        "Connection opened. But nested connections not allowed. Please, call Tmux with `t` key",
                        severity="warning",
                    )
                else:
                    self.notify(str(e), title="Internal", severity="error")
            finally:
                self.refresh()
                driver.start_application_mode()
        return result


def tui():
    SSHTui().run()


if __name__ == "__main__":
    tui()



import click

from .snippets import snippets_list, snippets_run


# ------------------------------------------------------------------------------
# Identity Commands
# ------------------------------------------------------------------------------
@click.group(name="snippets", help="Execute a Snippet")
def generate():
    pass


# // Linking other sub-commands
generate.add_command(snippets_run.cmd)
generate.add_command(snippets_list.cmd)



import click

from sshtmux.sshm import (
    SSH_Config,
    SSH_Group,
    complete_params,
    complete_ssh_group_names,
    complete_ssh_host_names,
)
from sshtmux.sshm.sshutils import validate_ssh_params, validate_unique_param

# ------------------------------------------------------------------------------
# COMMAND: host set
# ------------------------------------------------------------------------------
SHORT_HELP = "Set/Change host configuration"
LONG_HELP = """
Set/Change host definitions and parameters

If autocompletion is enabled, command will also offer current configured hosts for TAB completion.
"""

INFO_HELP = "Set host info, can be set multiple times, or set to empty value to clear it (example: -i '')"
SET_PARAM_HELP = "Sets parameter for the host, takes 2 values (<sshparam> <value>)"
UNSET_PARAM_HELP = "Unsets/Remove parameter for the host, takes 1 value <sshparam>"
GROUP_HELP = "Group change for host. Moves one or many hosts from their groups (they can be in different ones) to specified group."
FORCE_HELP = "Allows during group host group change, to move host(s) to group that does not exist, by creating it automatically, and moving specified hosts to new group."
# ------------------------------------------------------------------------------


@click.command(name="set", short_help=SHORT_HELP, help=LONG_HELP)
@click.option("-i", "--info", multiple=True, help=INFO_HELP)
@click.option(
    "-p",
    "--parameter",
    nargs=2,
    multiple=True,
    help=SET_PARAM_HELP,
    shell_complete=complete_params,
)
@click.option(
    "-r",
    "--remove-parameter",
    nargs=1,
    multiple=True,
    help=UNSET_PARAM_HELP,
    shell_complete=complete_params,
)
@click.option(
    "-g",
    "--group",
    "target_group_name",
    help=GROUP_HELP,
    shell_complete=complete_ssh_group_names,
)
@click.option("-f", "--force", is_flag=True, help=FORCE_HELP)
@click.argument("name", nargs=1, required=True, shell_complete=complete_ssh_host_names)
@click.pass_context
def cmd(ctx, name, info, parameter, remove_parameter, target_group_name, force):
    config: SSH_Config = ctx.obj

    if not config.check_host_by_name(name, False):
        print(
            f"Cannot set anything on host '{name}' as it is not defined in configuration!"
        )
        ctx.exit(1)

    found_host, found_group = config.get_host_by_name(name)

    if target_group_name:
        target_group_exists = config.check_group_by_name(target_group_name)
        if not target_group_exists and not force:
            print(
                f"Cannot move host '{name}' to group '{target_group_name}' which does not exist!"
            )
            print(
                "Consider using --force to automatically create target group, or create it manually first."
            )
            ctx.exit(1)
        elif not target_group_exists:
            target_group = SSH_Group(name=target_group_name)
            config.groups.append(target_group)
        else:
            target_group = config.get_group_by_name(target_group_name)

        config.move_host_to_group(found_host, found_group, target_group)

    if info:
        if len(info[0]) > 0:
            found_host.info = found_host.info + list(info)
        else:
            found_host.info = []

    valid_ssh_params = validate_ssh_params(parameter)
    validate_unique_param(found_host.params, valid_ssh_params)

    for param, value in valid_ssh_params.items():
        found_host.params[param] = value

    for param in remove_parameter:
        try:
            del found_host.params[param]
        except KeyError:
            print(f"Parameter: {param} not found to be removed. Ignoring...")

    if not config.stdout:
        print(f"Modified host: {name}")

    config.generate_ssh_config().write_out()



import click
from rich import print
from rich.prompt import Prompt

from sshtmux.exceptions import IdentityException
from sshtmux.services.identities import PasswordManager
from sshtmux.sshm import complete_identities

HELP = "Update Identity"


@click.command(name="update", short_help=HELP, help=HELP)
@click.argument("identity", shell_complete=complete_identities)
def cmd(identity):
    password_manager = PasswordManager()
    password = Prompt.ask("Type Password", password=True)
    try:
        password_manager.set_password(identity, password, True)
    except IdentityException as e:
        print(str(e))
        return

    print(f"Identity {identity} updated")



import click

from sshtmux.exceptions import IdentityException
from sshtmux.services.identities import PasswordManager
from sshtmux.sshm import complete_identities

HELP = "Delete Identity"


@click.command(name="delete", short_help=HELP, help=HELP)
@click.argument("identity", shell_complete=complete_identities)
def cmd(identity):
    password_manager = PasswordManager()
    try:
        password_manager.delete_password(identity)
    except IdentityException as e:
        print(str(e))
        return

    print(f"Identity {identity} deleted")



from .ssh_config import SSH_Config  # noqa: F401
from .ssh_group import SSH_Group  # noqa: F401
from .ssh_host import SSH_Host  # noqa: F401
from .ssh_parameters import *  # noqa: F401, F403
from .ssh_parameters import *  # noqa: F401, F403
from .ssh_graph import generate_graph  # noqa: F401
from .sshutils import (
    complete_params,  # noqa: F401
    complete_ssh_group_names,  # noqa: F401
    complete_ssh_host_names,  # noqa: F401
    complete_styles,  # noqa: F401
    expand_names,  # noqa: F401
    trace_jumphosts,  # noqa: F401
    complete_identities, # noqa: F401
)



import click

from .identity import (
    identity_create,
    identity_delete,
    identity_generate,
    identity_list,
    identity_run,
    identity_update,
)


# ------------------------------------------------------------------------------
# Identity Commands
# ------------------------------------------------------------------------------
@click.group(name="identity", help="Command group for managing identities")
def generate():
    pass


# // Linking other sub-commands
generate.add_command(identity_generate.cmd)
generate.add_command(identity_list.cmd)
generate.add_command(identity_create.cmd)
generate.add_command(identity_update.cmd)
generate.add_command(identity_delete.cmd)
generate.add_command(identity_run.cmd)



import click

from sshtmux.sshm import SSH_Config, SSH_Group, SSH_Host

# ------------------------------------------------------------------------------
# COMMAND: group create
# ------------------------------------------------------------------------------
SHORT_HELP = "Create new group"
LONG_HELP = """
Create new group

Group is used as visual and searchable separation between hosts.
It is achieved trough readable comments with special comment keyword
defining start of the group "#@group: <group-name>", every host defined
after this keyword is considered as part of this group.
"""

# Parameters help:
DESC_HELP = "Short description of group"
INFO_HELP = "Info line, can be set multiple times"
# ------------------------------------------------------------------------------


@click.command(name="create", short_help=SHORT_HELP, help=LONG_HELP)
@click.option("-d", "--desc", default="", help=DESC_HELP)
@click.option("-i", "--info", multiple=True, help=INFO_HELP)
@click.argument("name")
@click.pass_context
def cmd(ctx, name, desc, info):
    config: SSH_Config = ctx.obj

    # Check if already group exists
    if config.check_group_by_name(name):
        print(
            f"Cannot create new group '{name}', as group already exists with this name"
        )
        ctx.exit(1)

    new_group = SSH_Group(name, desc=desc, info=list(info))

    # Create group parttern host
    new_host = SSH_Host(name=f"{name}-*", group=name, type="pattern", info=[])
    new_group.patterns.append(new_host)

    # Add new group to config and show newly created group
    config.groups.append(new_group)
    config.generate_ssh_config().write_out()

    if not config.stdout:
        print(f"Created group: {name}")



import click
from libtmux.exc import LibTmuxException

from sshtmux.services.tmux import Tmux

HELP = "Show Snippets menu to execute one"


@click.command(name="run", short_help=HELP, help=HELP)
@click.option("--session-name", "-s", default=None, help="Session Name (opcional)")
@click.option("--window-index", "-w", default=None, help="Window Index (opcional)")
@click.option("--panel-index", "-p", default=None, help="Panel Index (opcional)")
def cmd(session_name, window_index, panel_index):
    tmux = Tmux()
    try:
        tmux.execute_snippet(session_name, window_index, panel_index)
    except LibTmuxException as e:
        print(str(e))



from pathlib import Path
from pprint import pprint
from typing import Tuple

import toml
from pydantic import ValidationError

from sshtmux.sshm import SSH_Config

from .core.config import (
    FAST_CONNECTIONS_GROUP_NAME,
    FAST_SESSIONS_NAME,
    MULTICOMMNAD_CLI,
    SFTP_CLI,
    ConfigModel,
    settings,
    update_settings,
)

Path(settings.internal_config.BASE_DIR).mkdir(parents=True, exist_ok=True)
Path(settings.sshtmux.SSHTMUX_SNIPPETS_PATH).mkdir(parents=True, exist_ok=True)


def init_toml_config():
    toml_file = settings.internal_config.TOML_CONFIG_FILE

    if Path(toml_file).exists():
        with open(toml_file, "r+t") as file:
            toml_dict = toml.load(file)
            try:
                if toml_dict.get("internal_config"):
                    del toml_dict["internal_config"]
                load_settings = ConfigModel(**toml_dict)
                update_settings(load_settings)
            except ValidationError as e:
                print("Invalid config.toml. Detail:", "\n")
                for error in e.errors():
                    msg = error.get("msg")
                    loc = error.get("loc")
                    if msg and loc:
                        if isinstance(loc, Tuple) and len(loc) == 2:
                            loc = loc[1]
                        print(f"Field: {loc}")
                        print(f"Error: {msg}")
                    else:
                        pprint(error, indent=4)
                exit(1)

    if not Path(toml_file).exists():
        with open(toml_file, "w+t") as file:
            toml_settings = settings.model_dump()
            if toml_settings.get("internal_config"):
                del toml_settings["internal_config"]
            toml.dump(toml_settings, file)


def init_tmux():
    if Path(settings.tmux.TMUX_CONFIG_FILE).exists():
        return

    tmux_config = """
    # Prefix
    set -g prefix C-b
    unbind C-b

    # Base
    set -g default-terminal "screen-256color"
    set -g visual-activity on
    set-option -g aggressive-resize on

    # Window
    set -g set-titles on
    set-option -g renumber-windows on
    set -g history-limit 100000
    set -g base-index 1
    setw -g pane-base-index 1
    bind '"' split-window -v -c "#{pane_current_path}"
    bind % split-window -h -c "#{pane_current_path}"
    bind-key -n MouseDown1StatusLeft choose-window

    # Mouse
    set -g mouse on
    set-option -g set-clipboard on

    # SSHTMUX Key Binds
    bind-key S run-shell "tmux split-window -h -c '#{pane_current_path}' 'sshm snippets run -s '#{session_name}' -w '#{window_index}' -p '#{pane_index}' '"
    bind-key I run-shell "tmux split-window -h -c '#{pane_current_path}' 'sshm identity run '#{session_name}' '#{window_index}' '#{pane_index}' '"
    bind-key F run-shell "tmux split-window -v -c '#{pane_current_path}' 'sshm host run '#{session_name}' '#{window_index}' '#{pane_index}' __SFTP_CLI__ '"
    bind-key M run-shell "tmux split-window -v -c '#{pane_current_path}' 'sshm host run '#{session_name}' '#{window_index}' '#{pane_index}' __MULTICOMMNAD_CLI__ '"

    # Key Binds useful
    bind-key -n M-s choose-session
    bind-key -n M-t choose-window
    bind-key -n M-n swap-window -t :- \\; select-window -t :-
    bind-key -n M-m swap-window -t :+ \\; select-window -t :+
    bind-key -n M-Left swap-window -t :- \\; select-window -t :-
    bind-key -n M-Right swap-window -t :+ \\; select-window -t :+
    bind-key -n M-d switch -t __DEFAULT_GROUP_NAME__
    bind-key -n M-f switch -t __FAST_CONNECTIONS_NAME__
    bind-key -n M-g switch -t __FAST_SESSIONS_NAME__
    bind-key -n M-r switch-client -n
    bind-key -n M-e switch-client -p
    bind-key -n M-o switch-client -p
    bind-key -n M-p switch-client -n
    bind-key -n M-q select-window -t :-
    bind-key -n M-w select-window -t :+1
    bind-key -n M-1 select-window -t 1
    bind-key -n M-2 select-window -t 2
    bind-key -n M-3 select-window -t 3
    bind-key -n M-4 select-window -t 4
    bind-key -n M-5 select-window -t 5
    bind-key -n M-6 select-window -t 6
    bind-key -n M-7 select-window -t 7
    bind-key -n M-8 select-window -t 8
    bind-key -n M-9 select-window -t 9

    # Copy Mouse Selection to clipboard
    # xclip=X11 | xsel=X11 | wl-cop=Wayland | pbcopy=MacOS
    if-shell 'command -v xclip >/dev/null' 'set-option -g @copy-command "xclip -selection clipboard -i"'
    if-shell 'command -v xsel >/dev/null' 'set-option -g @copy-command "xsel --clipboard -i"'
    if-shell 'command -v wl-copy >/dev/null' 'set-option -g @copy-command "wl-copy"'
    if-shell 'command -v pbcopy >/dev/null' 'set-option -g @copy-command "pbcopy"'
    bind-key -T copy-mode-vi MouseDragEnd1Pane send-keys -X copy-pipe-and-cancel "#{@copy-command}"
    bind-key -T copy-mode MouseDragEnd1Pane send-keys -X copy-pipe-and-cancel "#{@copy-command}"

    # Status Bar
    set -g status-interval 2
    set -g status-bg colour235
    set -g status-fg white
    set -g status-style bold
    set -g status-left-length 100
    set -g status-right-length 100
    set -g status-left '#[fg=white,bg=colour235,bold] 📚 #S #[default]'
    set -g status-right '#[fg=white]#{?#{==:#{pane_current_command},ssh},🔐 SSH,#{?#{==:#{pane_current_command},sftp},📂 SFTP,💻 Local}} #[fg=white]📅 %a %d/%b #[fg=white]🕒 %H:%M #[default]'

    # Tabs
    set -g window-status-current-style bg=colour33,fg=white,bold
    set -g window-status-current-format '#[fg=white,bold] #{?#{==:#{pane_current_command},ssh},🔐,#{?#{==:#{pane_current_command},sftp},📂,💻}} #I #W'
    set -g window-status-format '#[fg=colour250] #{?#{==:#{pane_current_command},ssh},🔐,#{?#{==:#{pane_current_command},sftp},📂,💻}} #I #W'
    set -g window-status-style bg=colour235,fg=colour250

    # Panes
    set -g pane-active-border-style fg=colour45
    set -g pane-border-style fg=colour235
    """

    tmux_config = (
        tmux_config.replace("__SFTP_CLI__", SFTP_CLI)
        .replace("__MULTICOMMNAD_CLI__", MULTICOMMNAD_CLI)
        .replace("__DEFAULT_GROUP_NAME__", SSH_Config.DEFAULT_GROUP_NAME)
        .replace("__FAST_CONNECTIONS_NAME__", FAST_CONNECTIONS_GROUP_NAME)
        .replace("__FAST_SESSIONS_NAME__", FAST_SESSIONS_NAME)
    )
    with open(settings.tmux.TMUX_CONFIG_FILE, "a+t") as file:
        file.write(tmux_config)


init_toml_config()
init_tmux()



from typing import List

from rich.padding import Padding
from rich.table import Table
from rich.text import Text

from sshtmux.sshm import SSH_Host

# Fixed width for jump-proxy cell in graph table (Keep value as multiple of 2!)
MAX_CELL_SIZE = 26
PADDING_LR = (0, 1)

HOST_NAME_COLOR = "bold white"
LOC_TUNNEL_COLOR = "cyan"
REM_TUNNEL_COLOR = "bold yellow"
ADDRESS_COLOR = {
    "online": "green",
    "offline": "red",
    "unchecked": "white",
}

# This are chars we use to draw...
LINK_UP = "│    │"
LINK_LR = "╯    ╰"
LINK_LR_EXT = "─"


# -----------------------------------------------------------------------------
# Graph generation function...
# -----------------------------------------------------------------------------
def generate_graph(traced_hosts: List[SSH_Host], print_tunnels=True):
    """
    Function that generates nice "graph" view of connected hosts
    """

    # TODO: This is currently static, but we could improve it, as in very long hostname
    #       it will result in broken lines, or alternatively in cut-off names
    reversed_hosts = traced_hosts[::-1]  # list of hosts in order of proxy-jumping

    # Create table with number of columns as we need to fit the data
    tbl = Table.grid()

    tbl.add_column(justify="right")  # source column
    for _ in range(len(traced_hosts) - 1):
        tbl.add_column(justify="center", width=MAX_CELL_SIZE)  # proxy-jump columns
    tbl.add_column(justify="left")  # target column

    # ----- Main graph row ----------------------------------------------------
    # Add Source info for main graph-row
    graph_row = [
        Padding(
            Text(
                "\n".join(
                    [
                        "Connection graph",
                        "────────────────",
                        "       THIS HOST",
                        "  SSH Connection",
                    ]
                )
            ),
            PADDING_LR,
        )
    ]

    # Add Jump proxies in graph-row
    for host in reversed_hosts[:-1]:
        # Status is filled by testing function before calling graph, so status is
        # already updated in given traced hosts
        host_status = host.params.get("status", "unchecked")
        address = f"[{ADDRESS_COLOR[host_status]}]{host.params.get('hostname', '')}[/]"

        graph_row.append(
            Padding(
                Text.from_markup(
                    "\n".join(
                        [
                            f"[{HOST_NAME_COLOR}]{host.name}[/]",
                            address,
                            LINK_UP,
                            LINK_LR.center(MAX_CELL_SIZE, LINK_LR_EXT),
                        ]
                    )
                )
            )
        )

    # Add Target info in graph-row
    target_status = reversed_hosts[-1].params.get("status", "unchecked")
    target_address = f"[{ADDRESS_COLOR[target_status]}]{reversed_hosts[-1].params.get('hostname', '')}[/]"

    graph_row.append(
        Padding(
            Text.from_markup(
                "\n".join(
                    [
                        "",
                        "",
                        "TARGET",
                        f"[{HOST_NAME_COLOR}]{reversed_hosts[-1].name}[/]",
                        target_address,
                    ]
                )
            ),
            PADDING_LR,
        )
    )

    # Add graph-row in the render table
    tbl.add_row(*graph_row)

    if print_tunnels:
        # ----- Local SSH tunnels row ---------------------------------------------
        # Print local forward tunnel between current host and target host
        if "localforward" in traced_hosts[0].params:
            tbl.add_row(*[Padding("Local SSH tunnels", PADDING_LR)])

            for tun in traced_hosts[0].params["localforward"]:
                tun_port, tun_end = tun.split()
                tunnel_row = [Padding(":" + tun_port, PADDING_LR)]
                for _ in range(len(traced_hosts) - 1):
                    tunnel_row.append(
                        Padding(f"[{LOC_TUNNEL_COLOR}]{'>' * MAX_CELL_SIZE}[/]")
                    )
                tunnel_row.append(Padding(tun_end, PADDING_LR))

                tbl.add_row(*tunnel_row)

        # ----- Remote SSH tunnels row --------------------------------------------
        # Print remote forward tunnel between current host and target host
        if "remoteforward" in traced_hosts[0].params:
            # trace.add_row(["" for _ in range(num_elements)])   # Empty row
            tbl.add_row(*[Padding("Remote SSH tunnels", PADDING_LR)])

            for tun in traced_hosts[0].params["remoteforward"]:
                tun_port, tun_end = tun.split()
                tunnel_row = [Padding(tun_end, PADDING_LR)]
                for _ in range(len(traced_hosts) - 1):
                    tunnel_row.append(Padding("<" * MAX_CELL_SIZE))
                tunnel_row.append(Padding(":" + tun_port, PADDING_LR))
                tbl.add_row(*tunnel_row)

    return tbl



import click

from sshtmux.sshm import SSH_Config, complete_ssh_group_names, expand_names

# ------------------------------------------------------------------------------
# COMMAND: group delete
# ------------------------------------------------------------------------------
SHORT_HELP = "Delete group (Will remove all hosts from group)"
LONG_HELP = """
Delete group

Command accepts single or multiple groups names to delete.
If autocompletion is enabled, command will also offer current configured groups for TAB completion.
Alternatively when group "NAME" is specified with "r:" prefix, then value is used as regex match
to find all group names that match given pattern, and will expand target group list accordingly.

Regex matches and direct or autocompleted names can be mixed.
Group deletion will be prompted for confirmation.
"""

# Parameters help:
YES_HELP = "Skip confirmation and assume 'yes'. Be careful!"
# ------------------------------------------------------------------------------


@click.command(name="delete", short_help=SHORT_HELP, help=LONG_HELP)
@click.option("--yes", is_flag=True, help=YES_HELP)
@click.argument(
    "names", nargs=-1, required=True, shell_complete=complete_ssh_group_names
)
@click.pass_context
def cmd(ctx, names, yes):
    config: SSH_Config = ctx.obj

    selected_group_list = expand_names(names, config.get_all_group_names())
    selected_group_list.sort()

    # Deleting requires confirmation
    if not yes:
        print(f"Following groups will be deleted: [{','.join(selected_group_list)}]")
        if not click.confirm("Are you sure?"):
            ctx.exit(1)

    # When deleting multiple groups, iterate over all of them
    config_updated = False
    for name in selected_group_list:
        # Find group by name
        if not config.check_group_by_name(name):
            print(f"Cannot delete group '{name}', it is not defined in configuration!")
            continue

        config.groups.remove(config.get_group_by_name(name))
        config_updated = True

        if not config.stdout:
            print(f"Deleted group: {name}")

    # ReWrite config only when config was actually changed
    if config_updated:
        config.generate_ssh_config().write_out()



import json
import os

from cryptography.fernet import Fernet, InvalidToken
from rich.console import Console
from rich.prompt import Prompt

from sshtmux.core.config import settings
from sshtmux.exceptions import IdentityException


class KeyManager:
    def __init__(self):
        self.key_file = settings.sshtmux.SSHTMUX_IDENTITY_KEY_FILE
        self.key_env = settings.sshtmux.SSHTMUX_IDENTITY_KEY
        if self.key_env and isinstance(self.key_env, str) and len(self.key_env) > 0:
            self.key = self.key_env
        else:
            self.key = self._load_or_generate_key()

    @staticmethod
    def generate_key():
        return Fernet.generate_key()

    def _load_or_generate_key(self):
        if os.path.exists(self.key_file):
            with open(self.key_file, "rb") as f:
                return f.read()
        else:
            key = KeyManager.generate_key()
            self._save_key(key)
            return key

    def _save_key(self, key):
        with open(self.key_file, "wb") as f:
            f.write(key)

    def get_key(self):
        return self.key


class PasswordManager:
    def __init__(self, service=settings.internal_config.BASE_SERVICE):
        self.service = service
        self.key_manager = KeyManager()
        self.password_file = settings.sshtmux.SSHTMUX_IDENTITY_PASSWORDS_FILE
        self.fernet = Fernet(self.key_manager.get_key())

    def _encrypt_data(self, data):
        try:
            return self.fernet.encrypt(data.encode())
        except Exception as e:
            raise IdentityException(str(e))

    def _decrypt_data(self, encrypted_data):
        try:
            return self.fernet.decrypt(encrypted_data).decode()
        except Exception as e:
            raise IdentityException(str(e))

    def _load_identities(self):
        if not os.path.exists(self.password_file):
            return {}
        with open(self.password_file, "rb") as f:
            encrypted_data = f.read()
        return json.loads(self._decrypt_data(encrypted_data))

    def _save_identities(self, passwords):
        encrypted_data = self._encrypt_data(json.dumps(passwords))
        with open(self.password_file, "wb") as f:
            f.write(encrypted_data)

    def set_password(self, reference, password, is_update=False):
        try:
            identities = self._load_identities()
        except InvalidToken:
            raise IdentityException("Invalid Token")
        except Exception as e:
            raise IdentityException(str(e))

        if not identities.get(self.service):
            identities[self.service] = {}
        identity = identities[self.service].get(reference)

        if not is_update:
            if identity:
                raise IdentityException("Identity already exists")
        identities[self.service][reference] = password

        self._save_identities(identities)

    def get_password(self, reference):
        identities = self._load_identities()
        users = identities.get(self.service)
        if not users:
            raise IdentityException("No Identities was Found")
        identity = users.get(reference)
        if not identity:
            raise IdentityException("Identity not Found")
        return identity

    def delete_password(self, reference):
        identities = self._load_identities()
        if not identities.get(self.service):
            return

        if reference in identities[self.service]:
            del identities[self.service][reference]
            self._save_identities(identities)
        else:
            raise IdentityException("User not Found")

    def get_identities(self):
        identities = self._load_identities()
        users = identities.get(self.service)
        if not users:
            return []
        return list(users.keys())


def prompt_identity():
    console = Console()
    password_manager = PasswordManager()
    identities = ["Cancel"] + password_manager.get_identities()
    identities_idx = [str(index) for index, _ in enumerate(identities)]

    console.print("Identities:", style="bold underline")
    for idx, option in enumerate(identities):
        console.print(f"{idx}. {option}")

    choice = Prompt.ask(
        "Choose Identity:",
        choices=identities_idx,
        default=identities_idx[0],
        show_choices=False,
    )
    identity = identities[int(choice)]
    if identity == identities[0]:
        return

    password = password_manager.get_password(identity)
    return password



import click

from sshtmux.sshm import SSH_Config, complete_ssh_host_names, expand_names

# ------------------------------------------------------------------------------
# COMMAND: host delete
# ------------------------------------------------------------------------------
SHORT_HELP = "Delete host(s)"
LONG_HELP = """
Delete host(s) from configuration

Command accepts single or multiple host names to delete.
If autocompletion is enabled, command will also offer current configured hosts for TAB completion.
Alternatively when "NAME" is specified with "r:" prefix, then part after ":" is used as regex match
to find all hosts that match that pattern, and will expand target host list accordingly.
If multiple regex matches are defined, results from all individual matches will be combined to target host list (exclusive OR)

\b
Regex NAMES example command: (sshm host delete r:^test_ r:_test$)
-> will delete all hosts which names start with "test_" or end with "_test"

Confirmation dialog will appear to confirm if deletion is ok to continue.
"""

# Parameters help:
YES_HELP = "Skip confirmation and assume 'yes'. Be careful!"
# ------------------------------------------------------------------------------


@click.command(name="delete", short_help=SHORT_HELP, help=LONG_HELP)
@click.option("--yes", is_flag=True, help=YES_HELP)
@click.argument(
    "names", nargs=-1, required=True, shell_complete=complete_ssh_host_names
)
@click.pass_context
def cmd(ctx, names, yes):
    config: SSH_Config = ctx.obj

    selected_hosts_list = expand_names(names, config.get_all_host_names())
    selected_hosts_list.sort()

    # Deleting requires confirmation
    if not yes:
        print(f"Following hosts will be deleted: [{','.join(selected_hosts_list)}]")
        if not click.confirm("Are you sure?"):
            ctx.exit(1)

    # When deleting multiple hosts, iterate over all of them
    for name in selected_hosts_list:
        if not config.check_host_by_name(name, False):
            print(f"Cannot delete host '{name}' as it is not defined in configuration!")
            continue

        found_host, found_group = config.get_host_by_name(name)

        if found_host.type == "normal":
            found_group.hosts.remove(found_host)
        else:
            found_group.patterns.remove(found_host)

        if not config.stdout:
            print(f"Deleted host: {name}")

    config.generate_ssh_config().write_out()
